import os
import sqlite3
from itertools import product

from tests.integration.basetests import BaseTestPycolator


class TestReassign(BaseTestPycolator):
    command = 'reassign'
    suffix = '_reassigned.xml'

    def test_reassign_psm(self):
        self.run_qvality('psm')

    def test_reassign_peptide(self):
        self.run_qvality('peptide')

    def run_qvality(self, feattype):
        def get_mapvals_interpolated(qpmap, svm):
            testsvm = float(svm)
            for map_svm in sorted([float(x) for x in qpmap], reverse=True):
                if testsvm < map_svm:
                    upperval = str(map_svm)
                elif testsvm > map_svm:
                    lowerval = str(map_svm)
                    return tuple([str(sum([float(qpmap[upperval][x]),
                                           float(qpmap[lowerval][x])]) / 2)
                                  for x in [0, 1]])
        qvalityfn = os.path.join(self.fixdir, 'qvality.txt')
        self.run_command(['--qvality', qvalityfn, '--feattype', feattype])
        result = self.read_percolator_out(self.resultfn)
        origin = self.read_percolator_out(self.infile[0])
        qvmap = {}
        with open(qvalityfn) as fp:
            next(fp)  # skip header
            for line in fp:
                line = line.strip('\n').split('\t')
                qvmap[line[0]] = (line[1], line[2])
        feat = '{}s'.format(feattype)
        id_id = '{}_id'.format(feattype)
        resultids = self.get_element_ids(result[feat], id_id, result['ns'])
        resultsvms = self.get_svms(result[feat], result['ns'])
        resultqvals = self.get_qvals(result[feat], result['ns'])
        resultpeps = self.get_peps(result[feat], result['ns'])
        originids = self.get_element_ids(origin[feat], id_id, origin['ns'])
        originsvms = self.get_svms(origin[feat], origin['ns'])
        for oid, rid, osvm, rsvm, rq, rpep in zip(originids, resultids,
                                                  originsvms, resultsvms,
                                                  resultqvals, resultpeps):
            self.assertEqual(oid, rid)
            self.assertEqual(osvm, rsvm)
            try:
                mapvals = qvmap[str(round(float(osvm), 5))]
            except KeyError:
                mapvals = get_mapvals_interpolated(qvmap, osvm)
            self.assertEqual(rq, mapvals[1])
            self.assertEqual(rpep, mapvals[0])


class TestSplit(BaseTestPycolator):
    def setUp(self):
        super().setUp()
        self.resultfn = None

    def do_check(self, first_exp_fn, sec_exp_fn, first_result, second_result):
        first_expected = os.path.join(self.fixdir, first_exp_fn)
        second_expected = os.path.join(self.fixdir, sec_exp_fn)
        first_exp_contents = self.read_percolator_out(first_expected)
        second_exp_contents = self.read_percolator_out(second_expected)
        first_contents = self.read_percolator_out(first_result)
        second_contents = self.read_percolator_out(second_result)

        self.assertEqual(len(first_contents['psms']),
                         len(first_exp_contents['psms']))
        self.assertEqual(len(first_contents['peptides']),
                         len(first_exp_contents['peptides']))
        self.assertEqual(len(second_contents['psms']),
                         len(second_exp_contents['psms']))
        self.assertEqual(len(second_contents['peptides']),
                         len(second_exp_contents['peptides']))
        return first_contents, second_contents


class TestSplitProteinHeader(TestSplit):
    command = 'splitprotein'
    suffix = ''

    def test_splitprotein(self):
        protheaders = ['ENSP', 'random;decoy']
        self.suffix = ''
        options = ['--protheaders'] + protheaders
        self.run_command(options)
        ensp_result = os.path.join(self.workdir, self.infilename + '_h0.xml')
        fake_result = os.path.join(self.workdir, self.infilename + '_h1.xml')
        ensp_contents, dr_contents = self.do_check('splitprotein_h0.xml',
                                                   'splitprotein_h1.xml',
                                                   ensp_result, fake_result)
        ns = self.get_xml_namespace(ensp_result)
        for feat in ['psms', 'peptides']:
            for el in ensp_contents[feat]:
                self.assertEqual(
                    el.find('{%s}protein_id' % ns['xmlns']).text[:4], 'ENSP')
            for el in dr_contents[feat]:
                self.assertIn(el.find('{%s}protein_id' % ns['xmlns']).text[:6],
                              ['decoy_', 'random'])


class TestSplitTD(TestSplit):
    command = 'splittd'
    suffix = ''

    def test_splittd(self):
        """TestSplitTD:: test_splittd
        Tests that splitted files contain equal amount of PSMS
        when compared with expected output, and checks that each psm/peptide
        has correct 'decoy' attribute."""
        target_result = os.path.join(self.workdir,
                                     self.infilename + '_target.xml')
        decoy_result = os.path.join(self.workdir,
                                    self.infilename + '_decoy.xml')
        self.run_command()
        t_contents, d_contents = self.do_check('splittd_target_out.xml',
                                               'splittd_decoy_out.xml',
                                               target_result, decoy_result)
        for feat in ['psms', 'peptides']:
            for el in t_contents[feat]:
                self.assertEqual(
                    el.attrib['{%s}decoy' % t_contents['ns']], 'false')
            for el in d_contents[feat]:
                self.assertEqual(
                    el.attrib['{%s}decoy' % d_contents['ns']], 'true')


class TestMerge(BaseTestPycolator):
    command = 'merge'
    infilename = 'splittd_target_out.xml'
    suffix = '_merged.xml'

    def test_merge_onefile(self):
        self.infile = [self.infile]
        self.run_command()
        expected = self.read_percolator_out(self.infile[0])
        result = self.read_percolator_out(self.resultfn)
        self.assertEqual(len(result['psms']), len(result['peptides']))
        self.assertCountEqual(self.get_element_ids(expected['psms'], 'psm_id',
                                                   expected['ns']),
                              self.get_element_ids(result['psms'], 'psm_id',
                                                   result['ns']))
        self.assertCountEqual(self.get_element_ids(expected['peptides'],
                                                   'peptide_id',
                                                   expected['ns']),
                              self.get_element_ids(result['peptides'],
                                                   'peptide_id', result['ns']))

    def test_merge(self):
        #self.multifiles = [os.path.join(self.fixdir, 'splittd_decoy_out.xml')]
        #self.multifiles =
        #options = ['--multifiles']
        #options.extend(self.multifiles)
        self.infile = [self.infile]
        self.infile.append(os.path.join(self.fixdir, 'splittd_decoy_out.xml'))
        self.run_command()
        expected = self.read_percolator_out(os.path.join(self.fixdir,
                                                         'percolator_out.xml'))
        result = self.read_percolator_out(self.resultfn)
        self.assertEqual(len(result['psms']), len(result['peptides']))
        self.assertCountEqual(self.get_element_ids(expected['psms'], 'psm_id',
                                                   expected['ns']),
                              self.get_element_ids(result['psms'], 'psm_id',
                                                   result['ns']))
        self.assertCountEqual(self.get_element_ids(expected['peptides'],
                                                   'peptide_id',
                                                   expected['ns']),
                              self.get_element_ids(result['peptides'],
                                                   'peptide_id', result['ns']))


class TestFilterUnique(BaseTestPycolator):
    command = 'filteruni'
    suffix = '_filtuni.xml'
    # FIXME other scores than svm
    # FIXME make sure BEST psm is retained, not worst.
    # FIXME illegal scores handling
    # FIXME PSM peptide reffing

    def test_filter_uniques(self):
        """Checks if resultpeps gets uniques, and also that input peptides
        were not unique to start with."""
        self.run_command(['--score', 'svm'])
        result = self.read_percolator_out(self.resultfn)
        origin = self.read_percolator_out(self.infile[0])
        resultpeps = self.get_element_ids(result['peptides'],
                                          'peptide_id', result['ns'])
        originpeps = self.get_element_ids(origin['peptides'],
                                          'peptide_id', origin['ns'])
        self.assertEqual(len({x for x in resultpeps}), len(resultpeps))
        self.assertNotEqual(len({x for x in originpeps}), len(originpeps))


class TestFilterLength(BaseTestPycolator):
    command = 'filterlen'
    suffix = '_filtlen.xml'
    # FIXME need to check maxlen minlen input?

    def length_correct(self, seqs, minlen, maxlen):
        for seq in seqs:
            seq = self.strip_modifications(seq)
            self.assertGreaterEqual(len(seq), minlen)
            self.assertLessEqual(len(seq), maxlen)

    def all_peps_in_output(self, original_seqs, minlen, maxlen, testresult):
        for feat in original_seqs:
            seq = self.strip_modifications(feat)
            if len(seq) <= minlen or len(seq) >= maxlen:
                continue
            self.assertIn(feat, testresult)

    def test_filterlen(self):
        maxlen = 20
        minlen = 10
        self.run_command(['--maxlen', str(maxlen), '--minlen', str(minlen)])
        result = self.get_psm_pep_ids_from_file(self.resultfn)
        origin = self.get_psm_pep_ids_from_file(self.infile[0])

        self.length_correct(result['peptide_ids'], minlen, maxlen)
        self.length_correct(result['psm_seqs'], minlen, maxlen)
        self.all_peps_in_output(origin['peptide_ids'], minlen, maxlen,
                                result['peptide_ids'])
        self.all_peps_in_output(origin['psm_seqs'], minlen, maxlen,
                                result['psm_seqs'])


class TestFilterKnown(BaseTestPycolator):
    command = 'filterseq'
    suffix = '_filtseq.xml'
    dbfn = 'known_peptide_lookup.sqlite'
    reversed_dbfn = 'rev_known_peptide_lookup.sqlite'

    def test_noflags(self):
        self.dbpath = os.path.join(self.fixdir, self.dbfn)
        self.assert_seqs_correct()

    def test_ntermwildcards(self):
        max_falloff = 12
        self.dbpath = os.path.join(self.fixdir, self.reversed_dbfn)
        self.assert_seqs_correct(['--insourcefrag', str(max_falloff)],
                                 'ntermfalloff', max_falloff)

    def test_deamidate(self):
        self.dbpath = os.path.join(self.fixdir, self.dbfn)
        self.assert_seqs_correct(['--deamidate'], 'deamidate')

    def deamidate(self, sequence):
        aa_possible = [(aa,) if aa != 'D' else ('D', 'N') for aa in sequence]
        return list(''.join(aa) for aa in product(*aa_possible))

    def assert_seqs_correct(self, flags=[], seqtype=None, max_falloff=False):
        """Does the actual testing"""
        options = ['--dbfile', self.dbpath]
        options.extend(flags)
        self.run_command(options)
        result = self.get_psm_pep_ids_from_file(self.resultfn)
        origin = self.get_psm_pep_ids_from_file(self.infile[0])
        for feattype in ['peptide_ids', 'psm_seqs']:
            original_seqs = origin[feattype]
            result_seqs = result[feattype]
            db = sqlite3.connect(self.dbpath)
            for oriseq in original_seqs:
                seq_dbcheck = self.strip_modifications(oriseq)
                if seqtype == 'deamidate':
                    testseqs = self.deamidate(seq_dbcheck)
                else:
                    testseqs = [seq_dbcheck]
                for seq in testseqs:
                    if self.seq_in_db(db, seq, seqtype, max_falloff):
                        self.assertNotIn(oriseq, result_seqs)
                        break
                    else:
                        self.assertIn(oriseq, result_seqs)


class TestQvality(BaseTestPycolator):
    command = 'qvality'
    suffix = '_qvalityout.txt'

    def test_qvality(self):
        self.infilename = 'splittd_target_out.xml'
        self.decoyfn = 'splittd_decoy_out.xml'
        self.decoy = os.path.join(self.fixdir, self.decoyfn)
        #options = ['--decoyfn', self.decoy, '--feattype', 'blaja', ]
        self.fail('pycolator qvality integration testing not implemented yet')



from app.lookups.sqlite.base import ResultLookupInterface


class PSMDB(ResultLookupInterface):
    def add_tables(self):
        self.create_tables(['psms', 'psmrows', 'peptide_sequences',
                            'proteins', 'protein_evidence', 'protein_seq',
                            'prot_desc', 'protein_psm', 'genes',
                            'associated_ids'])

    def store_proteins(self, proteins, evidence_lvls=False, sequences=False):
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO proteins(protein_acc) '
            'VALUES(?)', proteins)
        self.conn.commit()
        cursor = self.get_cursor()
        if evidence_lvls:
            cursor.executemany(
                'INSERT INTO protein_evidence(protein_acc, evidence_lvl) '
                'VALUES(?, ?)', evidence_lvls)
        if sequences:
            cursor.executemany(
                'INSERT INTO protein_seq(protein_acc, sequence) '
                'VALUES(?, ?)', sequences)
        self.conn.commit()
        self.index_column('proteins_index', 'proteins', 'protein_acc')
        self.index_column('evidence_index', 'protein_evidence', 'protein_acc')

    def store_descriptions(self, descriptions):
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO prot_desc(protein_acc, description) '
            'VALUES(?, ?)', descriptions)
        self.conn.commit()
        self.index_column('protdesc_index', 'prot_desc', 'protein_acc')

    def store_gene_and_associated_id(self, feats):
        genes = ((mapped['gene'], protein) for protein, mapped in feats.items())
        self.store_genes(genes)
        syms = ((mapped['symbol'], protein) for protein, mapped in feats.items())
        self.store_associated_ids(syms)
        descs = ((protein, mapped['desc']) for protein, mapped in feats.items())
        self.store_descriptions(descs)

    def store_genes(self, genes):
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO genes(gene_acc, protein_acc) VALUES(?, ?)', genes)
        self.conn.commit()
        self.index_column('gene_index', 'genes', 'protein_acc')

    def store_associated_ids(self, assoc_ids):
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO associated_ids(assoc_id, protein_acc) VALUES(?, ?)',
            assoc_ids)
        self.conn.commit()
        self.index_column('associd_index', 'associated_ids', 'protein_acc')

    def get_protein_gene_map(self):
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT p.protein_acc, g.gene_acc, aid.assoc_id, d.description '
            'FROM proteins AS p '
            'LEFT OUTER JOIN genes AS g ON p.protein_acc=g.protein_acc '
            'LEFT OUTER JOIN associated_ids AS aid '
            'ON p.protein_acc=aid.protein_acc '
            'LEFT OUTER JOIN prot_desc AS d ON p.protein_acc=d.protein_acc'
        )
        gpmap = {p_acc: {'gene': gene, 'symbol': sym, 'desc': desc}
                 for p_acc, gene, sym, desc in cursor}
        return gpmap

    def store_pepseqs(self, sequences):
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO peptide_sequences(sequence) VALUES(?)', sequences)
        self.conn.commit()

    def store_psms(self, psms):
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO psms(psm_id, pep_id, score, spectra_id) '
            'VALUES(?, ?, ?, ?)', ((psm['psm_id'], psm['seq'], psm['score'],
                                    psm['spec_id'])
                                   for psm in psms))
        cursor.executemany(
            'INSERT INTO psmrows(psm_id, rownr) VALUES(?, ?)',
            ((psm['psm_id'], psm['rownr']) for psm in psms))
        self.conn.commit()

    def store_peptides_proteins(self, allpepprot, psmids_to_store):
        ppmap = {psm_id: allpepprot[psm_id] for psm_id in psmids_to_store}
        prot_psm_ids = ((prot_acc, psm_id)
                        for psm_id, prots in ppmap.items()
                        for prot_acc in prots)
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO protein_psm(protein_acc, psm_id)'
            ' VALUES (?, ?)', prot_psm_ids)
        self.conn.commit()

    def get_peptide_seq_map(self):
        cursor = self.get_cursor()
        seqs = cursor.execute('SELECT pep_id, sequence FROM peptide_sequences')
        return {seq: pepid for pepid, seq in seqs.fetchall()}

    def index_psms(self):
        self.index_column('psmid_index', 'psms', 'psm_id')
        self.index_column('psmspecid_index', 'psms', 'spectra_id')
        self.index_column('psmrowid_index', 'psmrows', 'psm_id')
        self.index_column('psmrow_index', 'psmrows', 'rownr')
        self.index_column('pepseq_index', 'peptide_sequences', 'sequence')
        self.index_column('pepid_index', 'peptide_sequences', 'pep_id')
        self.index_column('psmspepid_index', 'psms', 'pep_id')

    def index_protein_peptides(self):
        self.index_column('protein_index', 'protein_psm', 'protein_acc')
        self.index_column('protpsmid_index', 'protein_psm', 'psm_id')




from app.dataformats import peptable as peptabledata
from app.actions.shared.pepprot_isoquant import base_add_isoquant_data


def add_isoquant_data(peptides, quantpeptides, quantacc, quantfields):
    """Runs through a peptide table and adds quant data from ANOTHER peptide 
    table that contains that data."""
    for peptide in base_add_isoquant_data(peptides, quantpeptides, 
                                          peptabledata.HEADER_PEPTIDE,
                                          quantacc, quantfields):
        yield peptide 



import os
from app.readers import xml as basereader


def specfn_quant_generator(specfiles, quantfiles, tag, ignore_tags):
    """Generates tuples of specfile and quant element for general formats"""
    for specfn, qfn in zip(specfiles, quantfiles):
        for quant_el in basereader.generate_xmltags(qfn, tag, ignore_tags):
            yield os.path.basename(specfn), quant_el


def mzmlfn_cons_el_generator(specfiles, consfiles):
    """Returns generation of tuples of spectra file and
    consensusXML quant elements"""
    return specfn_quant_generator(specfiles, consfiles, 'consensusElement',
                                  ['consensusElementList'])


def mzmlfn_feature_generator(specfiles, featfiles):
    """Returns tuple of spectrafile and features of OpenMS
    feature XML format"""
    return specfn_quant_generator(specfiles, featfiles, 'feature',
                                  ['featureList'])


def get_consxml_rt(cons_el):
    """Returns consensusXML in seconds"""
    return cons_el.find('centroid').attrib['rt']


def get_feature_info(feature):
    """Returns a dict with feature information"""
    dimensions = feature.findall('position')
    for dim in dimensions:
        if dim.attrib['dim'] == '0':
            rt = dim.text
        elif dim.attrib['dim'] == '1':
            mz = dim.text
    return {'rt': float(rt), 'mz': float(mz),
            'charge': int(feature.find('charge').text),
            'intensity': float(feature.find('intensity').text),
            }


def get_quantmap(consfile):
    """Returns map of isobaric quant channel names as keys, channel numbers
    values from consensus XML.
    E.g. {'113': '0',
          '114': '1',}
    """
    quantmap = {}
    maplist = basereader.get_element(consfile, 'mapList')
    for mapitem in maplist:
        quantmap[mapitem.attrib['label']] = mapitem.attrib['id']
    return quantmap



from app.actions.mzidtsv import proteingrouping as prep
from app.drivers.mzidtsv import MzidTSVDriver
from app.drivers.options import mzidtsv_options


class ProteinGroupDriver(MzidTSVDriver):
    outsuffix = '_protgroups.txt'
    lookuptype = 'proteingroups'
    command = 'proteingroup'
    commandhelp = ('Takes lookup SQLite result, uses it to output '
                   'mzidtsv file with protein groups')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['lookupfn', 'unroll'],
                                                mzidtsv_options))

    def get_psms(self):
        self.header = prep.get_header_with_proteingroups(self.oldheader)
        self.psms = prep.generate_psms_with_proteingroups(self.fn,
                                                          self.oldheader,
                                                          self.header,
                                                          self.lookup,
                                                          self.unroll)



def write_tsv(headerfields, features, outfn):
    """Writes header and generator of lines to tab separated file.

    headerfields - list of field names in header in correct order
    features - generates 1 list per line that belong to header
    outfn - filename to output to. Overwritten if exists
    """
    with open(outfn, 'w') as fp:
        write_tsv_line_from_list(headerfields, fp)
        for line in features:
            write_tsv_line_from_list([str(line[field]) for field
                                      in headerfields], fp)


def write_tsv_line_from_list(linelist, outfp):
    """Utility method to convert list to tsv line with carriage return"""
    line = '\t'.join(linelist)
    outfp.write(line)
    outfp.write('\n')


def parse_NA(feature, header):
    for field in header:
        try:
            feature[field] = str(feature[field])
        except KeyError:
            feature[field] = 'NA'
    return feature



from app.lookups.sqlite.protpeptable import ProtPepTable


class PepTableProteinCentricDB(ProtPepTable):
    datatype = 'peptide'
    colmap = {'peptide_sequences': ['pep_id', 'sequence'],
              'peptide_precur_quanted': ['pep_id', 'peptable_id', 'quant'],
              'peptide_fdr': ['pep_id', 'peptable_id', 'fdr'],
              'peptide_pep': ['pep_id', 'peptable_id', 'pep'],
              'pepquant_channels': ['channel_id', 'peptable_id',
                                    'channel_name', 'amount_psms_name'],
              'peptide_iso_quanted': ['peptidequant_id', 'pep_id',
                                      'channel_id', 'quantvalue',
                                      'amount_psms'],
              }

    def add_tables(self):
        self.create_tables(['peptide_tables', 'pepquant_channels',
                            'peptide_iso_quanted', 'peptide_precur_quanted',
                            'peptide_fdr', 'peptide_pep'])


class PepTableGeneCentricDB(PepTableProteinCentricDB):
    datatype = 'peptide'

    def get_protein_gene_symbol_for_map(self):
        fields = ['p.gene_acc', 'pep.sequence', 'pd.description', 'aid.assoc_id']
        sql = (
                'SELECT {} FROM {} AS p '
                'JOIN protein_psm USING(protein_acc) '
                'JOIN psms USING(psm_id) '
                'JOIN peptide_sequences AS pep USING(pep_id) '
                'LEFT OUTER JOIN prot_desc AS pd USING(protein_acc) '
                'LEFT OUTER JOIN associated_ids AS aid USING(protein_acc)'
                )
        sql = sql.format(','.join(fields), 'genes')
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_proteins_psms_for_map(self):
        """Gets gene-PSM combinations from DB and filters out uniques
        on the fly. Filtering is done since PSM are stored per protein,
        not per gene, so there may be a lot of *plicates"""
        fields = ['p.gene_acc', 'sets.set_name',
                  'pep.sequence', 'psm.psm_id']
        firstjoin = ('protein_psm', 'pp', 'protein_acc')
        genetable = 'genes'
        return self.get_unique_gene_psms(genetable, fields, firstjoin)


class PepTablePlainDB(PepTableProteinCentricDB):
    datatype = 'peptide'

    def get_proteins_psms_for_map(self):
        fields = ['p.protein_acc', 'sets.set_name', 'pep.sequence',
                  'psm.psm_id']
        firstjoin = ('protein_psm', 'pp', 'protein_acc')
        return self.get_proteins_psms('proteins', fields, firstjoin)



import os

from app.actions.peptable import merge as preparation
from app.drivers.peptable.base import PeptableMergeDriver
from app.drivers.options import peptable_options


class BuildPeptideTableDriver(PeptableMergeDriver):
    outsuffix = ''
    lookuptype = 'peptidetable'
    command = 'build'
    commandhelp = ('Create peptide table from a lookup database. '
                   'E.g. when multiple peptide quant tables have '
                   'been read into the lookup and will be combined.'
                   'Does not need an -i.')

    def parse_input(self, **kwargs):
        """Build peptide table has no input file (though it has a lookup),
        which is why we set it to outfile name so the infile fetching
        and outfile creating wont error."""
        super().parse_input(**kwargs)
        self.fn = os.path.join(os.getcwd(), 'built_peptide_table.txt')
        if self.genecentric:
            self.lookuptype = 'peptidegenecentrictable'
        elif self.noncentric:
            self.genecentric = 'plain'
            self.lookuptype = 'peptidetableplain'

    def set_options(self):
        super().set_options()
        options = self.define_options(['lookupfn', 'genecentric', 'noncentric',
                                       'isobaric', 'precursor', 'fdr', 'pep',
                                       'mock_infn'], peptable_options)
        self.options.update(options)

    def set_feature_generator(self):
        """Generates proteins with quant from the lookup table"""
        self.features = preparation.build_peptidetable(self.lookup,
                                                       self.headerfields,
                                                       self.isobaric,
                                                       self.precursor,
                                                       self.fdr, self.pep,
                                                       self.genecentric)



from lxml import etree


def write_percolator_xml(staticxml, feats, fn):
    """Given the static percolator xml root and process info nodes, and all
    psms and peptides as iterators in a dict {'peptide': pep_iterator, 'psm':
    psm_iterator}, this generates percolator out data into a file."""

    # First get xml until psms opening element is found.
    etree.SubElement(staticxml, 'psms').text = '***psms***'
    root = etree.tostring(staticxml, pretty_print=True,
                          xml_declaration=True, encoding='UTF-8')
    root = root.decode('utf-8')
    root = root[:root.find('***psms***')]

    # Write opening xml
    with open(fn, 'w') as fp:
        fp.write(root)
        fp.write('\n')

    # Then write features
    with open(fn, 'a') as fp:
        psmcount = 0
        for psm in feats['psm']:
            psmcount += 1
            fp.write(psm)
            fp.write('\n')
        fp.write('</psms><peptides>\n')

        peptidecount = 0
        for pep in feats['peptide']:
            peptidecount += 1
            fp.write(pep)
            fp.write('\n')
        fp.write('</peptides></percolator_output>')
    print('Wrote {0} psms, {1} peptides to file {2}'.format(psmcount,
                                                            peptidecount, fn))


def write_qvality_input(scores, fn):
    with open(fn, 'w') as fp:
        for score in scores:
            fp.write('{0}\n'.format(score))



import sqlite3


mslookup_tables = {'biosets': ['set_id INTEGER PRIMARY KEY',
                               'set_name TEXT'],
                   'mzmlfiles': ['mzmlfile_id INTEGER PRIMARY KEY',
                                 'mzmlfilename TEXT',
                                 'set_id INTEGER',
                                 'FOREIGN KEY(set_id)'
                                 'REFERENCES biosets'
                                 ],
                   'mzml': ['spectra_id TEXT PRIMARY KEY',
                            'mzmlfile_id INTEGER',
                            'scan_nr TEXT',
                            'charge INTEGER',
                            'mz DOUBLE',
                            'retention_time DOUBLE',
                            'ion_injection_time DOUBLE',
                            'FOREIGN KEY(mzmlfile_id)'
                            'REFERENCES mzmlfiles'],
                   'isobaric_channels': ['channel_id INTEGER PRIMARY KEY',
                                         'channel_name TEXT'],
                   'isobaric_quant': ['spectra_id TEXT',
                                      'channel_id INTEGER',
                                      'intensity REAL',
                                      'FOREIGN KEY(spectra_id)'
                                      'REFERENCES mzml',
                                      'FOREIGN KEY(channel_id)'
                                      'REFERENCES isobaric_channels'
                                      ],
                   # ms1_quant has no spectra_id reference since it contains
                   # features and Im not sure if they can be linked to
                   # spectra_ids or that retention time is averaged
                   'ms1_quant': ['feature_id INTEGER PRIMARY KEY',
                                 'mzmlfile_id INTEGER',
                                 'retention_time REAL', 'mz REAL',
                                 'charge INTEGER', 'intensity REAL',
                                 'FOREIGN KEY(mzmlfile_id)'
                                 'REFERENCES mzmlfiles'],
                   'ms1_align': ['spectra_id TEXT',
                                 'feature_id INTEGER',
                                 'FOREIGN KEY(spectra_id) '
                                 'REFERENCES mzml '
                                 'FOREIGN KEY(feature_id) '
                                 'REFERENCES ms1_quant',
                                 ],
                   'peptide_sequences': ['pep_id INTEGER PRIMARY KEY',
                                         'sequence TEXT',
                                         ],
                   'psms': ['psm_id TEXT PRIMARY KEY NOT NULL',
                            'pep_id INTEGER',
                            'score TEXT',
                            'spectra_id TEXT',
                            'FOREIGN KEY(pep_id)'
                            'REFERENCES peptide_sequences '
                            'FOREIGN KEY(spectra_id)'
                            'REFERENCES mzml'
                            ],
                   'psmrows': ['psm_id TEXT',
                               'rownr INTEGER',
                               'FOREIGN KEY(psm_id) '
                               'REFERENCES psms(psm_id)'],
                   'proteins': ['pacc_id INTEGER PRIMARY KEY',
                                'protein_acc TEXT UNIQUE'],
                   'gene_tables': ['genetable_id INTEGER PRIMARY KEY',
                                   'set_id INTEGER',
                                   'filename TEXT',
                                   'FOREIGN KEY(set_id)'
                                   'REFERENCES biosets'],
                   'protein_tables': ['prottable_id INTEGER PRIMARY KEY',
                                      'set_id INTEGER',
                                      'filename TEXT',
                                      'FOREIGN KEY(set_id)'
                                      'REFERENCES biosets'],
                   'peptide_tables': ['peptable_id INTEGER PRIMARY KEY',
                                      'set_id INTEGER',
                                      'filename TEXT',
                                      'FOREIGN KEY(set_id)'
                                      'REFERENCES biosets'],
                   'pepquant_channels': ['channel_id INTEGER PRIMARY KEY',
                                         'peptable_id INTEGER',
                                         'channel_name TEXT',
                                         'amount_psms_name TEXT',
                                         'FOREIGN KEY(peptable_id) '
                                         'REFERENCES '
                                         'peptide_tables(peptable_id)'
                                         ],
                   'peptide_iso_quanted': ['peptidequant_id'
                                           'INTEGER PRIMARY KEY',
                                           'pep_id INTEGER',
                                           'channel_id INTEGER',
                                           'quantvalue REAL',
                                           'amount_psms INTEGER',
                                           'FOREIGN KEY(pep_id) '
                                           'REFERENCES '
                                           'peptide_sequences(pep_id) '
                                           'FOREIGN KEY(channel_id) '
                                           'REFERENCES '
                                           'pepquant_channels(channel_id)'
                                           ],
                   'peptide_precur_quanted':
                   ['pep_precquant_id INTEGER PRIMARY KEY',
                    'pep_id INTEGER',
                    'peptable_id INTEGER',
                    'quant REAL',
                    'FOREIGN KEY(pep_id) '
                    'REFERENCES peptide_sequences(pep_id) '
                    'FOREIGN KEY(peptable_id) '
                    'REFERENCES peptide_tables(peptable_id)'
                    ],
                   'peptide_fdr': ['pep_id INTEGER',
                                   'peptable_id INTEGER',
                                   'fdr DOUBLE',
                                   'FOREIGN KEY(pep_id) '
                                   'REFERENCES peptide_sequences(pep_id) '
                                   'FOREIGN KEY(peptable_id) '
                                   'REFERENCES '
                                   'peptide_tables(peptable_id)'
                                   ],
                   'peptide_pep': ['pep_id INTEGER',
                                   'peptable_id INTEGER',
                                   'pep DOUBLE',
                                   'FOREIGN KEY(pep_id) '
                                   'REFERENCES peptide_sequences(pep_id) '
                                   'FOREIGN KEY(peptable_id) '
                                   'REFERENCES '
                                   'peptide_tables(peptable_id)'
                                   ],
                   'protein_precur_quanted':
                   ['prot_precquant_id INTEGER PRIMARY KEY',
                    'pacc_id INTEGER',
                    'prottable_id INTEGER',
                    'quant REAL',
                    'FOREIGN KEY(pacc_id) '
                    'REFERENCES proteins(pacc_id) '
                    'FOREIGN KEY(prottable_id) '
                    'REFERENCES protein_tables(prottable_id)'
                    ],
                   'gene_precur_quanted':
                   ['gene_precquant_id INTEGER PRIMARY KEY',
                    'gene_id INTEGER',
                    'genetable_id INTEGER',
                    'quant REAL',
                    'FOREIGN KEY(gene_id) '
                    'REFERENCES genes(gene_id) '
                    'FOREIGN KEY(genetable_id) '
                    'REFERENCES gene_tables(genetable_id)'
                    ],
                   'assoc_precur_quanted':
                   ['gene_precquant_id INTEGER PRIMARY KEY',
                    'gene_id INTEGER',
                    'genetable_id INTEGER',
                    'quant REAL',
                    'FOREIGN KEY(gene_id) '
                    'REFERENCES associated_ids(gene_id) '
                    'FOREIGN KEY(genetable_id) '
                    'REFERENCES gene_tables(genetable_id)'
                    ],
                   'protein_iso_quanted': ['proteinquant_id '
                                           'INTEGER PRIMARY KEY',
                                           'pacc_id INTEGER',
                                           'channel_id INTEGER',
                                           'quantvalue REAL',
                                           'amount_psms INTEGER',
                                           'FOREIGN KEY(pacc_id) '
                                           'REFERENCES proteins(pacc_id) '
                                           'FOREIGN KEY(channel_id) '
                                           'REFERENCES '
                                           'protquant_channels(channel_id)'
                                           ],
                   'gene_iso_quanted': ['genequant_id '
                                        'INTEGER PRIMARY KEY',
                                        'gene_id INTEGER',
                                        'channel_id INTEGER',
                                        'quantvalue REAL',
                                        'amount_psms INTEGER',
                                        'FOREIGN KEY(gene_id) '
                                        'REFERENCES genes(gene_id) '
                                        'FOREIGN KEY(channel_id) '
                                        'REFERENCES '
                                        'genequant_channels(channel_id)'
                                        ],
                   'assoc_iso_quanted': ['genequant_id '
                                         'INTEGER PRIMARY KEY',
                                         'gene_id INTEGER',
                                         'channel_id INTEGER',
                                         'quantvalue REAL',
                                         'amount_psms INTEGER',
                                         'FOREIGN KEY(gene_id) '
                                         'REFERENCES associated_ids(gene_id) '
                                         'FOREIGN KEY(channel_id) '
                                         'REFERENCES '
                                         'genequant_channels(channel_id)'
                                         ],
                   'genequant_channels': ['channel_id INTEGER PRIMARY KEY',
                                          'genetable_id INTEGER',
                                          'channel_name TEXT',
                                          'amount_psms_name TEXT',
                                          'FOREIGN KEY(genetable_id) '
                                          'REFERENCES '
                                          'gene_tables(genetable_id)'
                                          ],
                   'protquant_channels': ['channel_id INTEGER PRIMARY KEY',
                                          'prottable_id INTEGER',
                                          'channel_name TEXT',
                                          'amount_psms_name TEXT',
                                          'FOREIGN KEY(prottable_id) '
                                          'REFERENCES '
                                          'protein_tables(prottable_id)'
                                          ],
                   'gene_probability': ['gene_id INTEGER',
                                        'genetable_id INTEGER',
                                        'probability DOUBLE',
                                        'FOREIGN KEY(gene_id) '
                                        'REFERENCES genes(gene_id) '
                                        'FOREIGN KEY(genetable_id) '
                                        'REFERENCES '
                                        'gene_tables(genetable_id)'
                                        ],
                   'assoc_probability': ['gene_id INTEGER',
                                         'genetable_id INTEGER',
                                         'probability DOUBLE',
                                         'FOREIGN KEY(gene_id) '
                                         'REFERENCES associated_ids(gene_id) '
                                         'FOREIGN KEY(genetable_id) '
                                         'REFERENCES '
                                         'gene_tables(genetable_id)'
                                         ],
                   'protein_probability': ['pacc_id INTEGER',
                                           'prottable_id INTEGER',
                                           'probability DOUBLE',
                                           'FOREIGN KEY(pacc_id) '
                                           'REFERENCES proteins(pacc_id) '
                                           'FOREIGN KEY(prottable_id) '
                                           'REFERENCES '
                                           'protein_tables(prottable_id)'
                                           ],
                   'gene_fdr': ['gene_id INTEGER',
                                'genetable_id INTEGER',
                                'fdr DOUBLE',
                                'FOREIGN KEY(gene_id) '
                                'REFERENCES genes(gene_id) '
                                'FOREIGN KEY(genetable_id) '
                                'REFERENCES '
                                'gene_tables(genetable_id)'
                                ],
                   'assoc_fdr': ['gene_id INTEGER',
                                 'genetable_id INTEGER',
                                 'fdr DOUBLE',
                                 'FOREIGN KEY(gene_id) '
                                 'REFERENCES associated_ids(gene_id) '
                                 'FOREIGN KEY(genetable_id) '
                                 'REFERENCES '
                                 'gene_tables(genetable_id)'
                                 ],
                   'protein_fdr': ['pacc_id INTEGER',
                                   'prottable_id INTEGER',
                                   'fdr DOUBLE',
                                   'FOREIGN KEY(pacc_id) '
                                   'REFERENCES proteins(pacc_id) '
                                   'FOREIGN KEY(prottable_id) '
                                   'REFERENCES '
                                   'protein_tables(prottable_id)'
                                   ],
                   'gene_pep': ['gene_id INTEGER',
                                'genetable_id INTEGER',
                                'pep DOUBLE',
                                'FOREIGN KEY(gene_id) '
                                'REFERENCES genes(gene_id) '
                                'FOREIGN KEY(genetable_id) '
                                'REFERENCES '
                                'gene_tables(genetable_id)'
                                ],
                   'assoc_pep': ['gene_id INTEGER',
                                 'genetable_id INTEGER',
                                 'pep DOUBLE',
                                 'FOREIGN KEY(gene_id) '
                                 'REFERENCES associated_ids(gene_id) '
                                 'FOREIGN KEY(genetable_id) '
                                 'REFERENCES '
                                 'gene_tables(genetable_id)'
                                 ],
                   'protein_pep': ['pacc_id INTEGER',
                                   'prottable_id INTEGER',
                                   'pep DOUBLE',
                                   'FOREIGN KEY(pacc_id) '
                                   'REFERENCES proteins(pacc_id) '
                                   'FOREIGN KEY(prottable_id) '
                                   'REFERENCES '
                                   'protein_tables(prottable_id)'
                                   ],
                   'protein_psm': ['protein_acc TEXT',
                                   'psm_id TEXT',
                                   'FOREIGN KEY(protein_acc) '
                                   'REFERENCES proteins(protein_acc) '
                                   'FOREIGN KEY(psm_id) '
                                   'REFERENCES psms(psm_id)'],
                   'protein_evidence': ['protein_acc TEXT',
                                        'evidence_lvl REAL',
                                        'FOREIGN KEY(protein_acc) '
                                        'REFERENCES '
                                        'proteins(protein_acc)'],
                   'protein_seq': ['protein_acc TEXT',
                                   'sequence TEXT',
                                   'FOREIGN KEY(protein_acc) '
                                   'REFERENCES '
                                   'proteins(protein_acc)'],
                   'protein_coverage': ['protein_acc TEXT',
                                        'coverage REAL',
                                        'FOREIGN KEY(protein_acc) '
                                        'REFERENCES '
                                        'proteins(protein_acc)'],
                   'protein_group_master': ['master_id INTEGER PRIMARY KEY',
                                            'protein_acc TEXT',
                                            'FOREIGN KEY(protein_acc) '
                                            'REFERENCES '
                                            'proteins(protein_acc)'],
                   'protein_group_content': ['protein_acc TEXT',
                                             'master_id INTEGER',
                                             'peptide_count INTEGER',
                                             'psm_count INTEGER',
                                             'protein_score INTEGER',
                                             'FOREIGN KEY(protein_acc) '
                                             'REFERENCES '
                                             'proteins(protein_acc) '
                                             'FOREIGN KEY(master_id) '
                                             'REFERENCES '
                                             'protein_group_master'
                                             '(master_id)'
                                             ],
                   'psm_protein_groups': ['psm_id TEXT',
                                          'master_id INTEGER',
                                          'FOREIGN KEY(psm_id) REFERENCES'
                                          ' psms(psm_id)',
                                          'FOREIGN KEY(master_id) REFERENCES'
                                          ' protein_group_master(master_id)'],
                   'genes': ['gene_id INTEGER PRIMARY KEY',
                             'gene_acc TEXT',
                             'protein_acc TEXT',
                             'FOREIGN KEY(protein_acc) '
                             'REFERENCES proteins(protein_acc)'],
                   'associated_ids': ['gene_id INTEGER PRIMARY KEY',
                                      'assoc_id TEXT',
                                      'protein_acc TEXT',
                                      'FOREIGN KEY(protein_acc) '
                                      'REFERENCES proteins(protein_acc)'],
                   'prot_desc': ['protein_acc TEXT',
                                 'description TEXT',
                                 'FOREIGN KEY(protein_acc) '
                                 'REFERENCES proteins(protein_acc)'],
                   'known_searchspace': ['seqs TEXT'],
                   'protein_peptides': ['seq TEXT', 'protid TEXT',
                                        'pos INTEGER'],
                   }


class DatabaseConnection(object):
    def __init__(self, fn=None):
        """SQLite connecting when given filename"""
        self.fn = fn
        if self.fn is not None:
            self.connect(self.fn)

    def get_fn(self):
        """Returns lookup filename"""
        return self.fn

    def create_tables(self, tables):
        """Creates database tables in sqlite lookup db"""
        cursor = self.get_cursor()
        for table in tables:
            columns = mslookup_tables[table]
            try:
                cursor.execute('CREATE TABLE {0}({1})'.format(
                    table, ', '.join(columns)))
            except sqlite3.OperationalError as error:
                print(error)
                print('Warning: Table {} already exists in database, will '
                      'add to existing tables instead of creating '
                      'new.'.format(table))
            else:
                self.conn.commit()

    def connect(self, fn):
        """SQLite connect method initialize db"""
        self.conn = sqlite3.connect(fn)
        cur = self.get_cursor()
        cur.execute('PRAGMA page_size=4096')
        cur.execute('PRAGMA FOREIGN_KEYS=ON')
        cur.execute('PRAGMA cache_size=10000')
        cur.execute('PRAGMA journal_mode=MEMORY')

    def get_cursor(self):
        """Quickly get cursor, abstracting connection"""
        return self.conn.cursor()

    def close_connection(self):
        """Close connection to db, abstracts connection object"""
        self.conn.close()

    def index_column(self, index_name, table, column):
        """Called by interfaces to index specific column in table"""
        cursor = self.get_cursor()
        try:
            cursor.execute(
                'CREATE INDEX {0} on {1}({2})'.format(index_name, table, column))
        except sqlite3.OperationalError as error:
            print(error)
            print('Skipping index creation and assuming it exists already')
        else:
            self.conn.commit()

    def get_inclause(self, inlist):
        """Returns SQL IN clauses"""
        return 'IN ({0})'.format(', '.join('?' * len(inlist)))

    def get_sql_select(self, columns, table, distinct=False):
        """Creates and returns an SQL SELECT statement"""
        sql = 'SELECT {0} {1} FROM {2}'
        dist = {True: 'DISTINCT', False: ''}[distinct]
        return sql.format(dist, ', '.join(columns), table)

    def store_many(self, sql, values):
        """Abstraction over executemany method"""
        cursor = self.get_cursor()
        cursor.executemany(sql, values)
        self.conn.commit()

    def execute_sql(self, sql):
        """Executes SQL and returns cursor for it"""
        cursor = self.get_cursor()
        cursor.execute(sql)
        return cursor


class ResultLookupInterface(DatabaseConnection):
    """Connection subclass shared by result lookup interfaces that need
    to query the database when storing to get e.g. spectra id"""

    def get_mzmlfile_map(self):
        """Returns dict of mzmlfilenames and their db ids"""
        cursor = self.get_cursor()
        cursor.execute('SELECT mzmlfile_id, mzmlfilename FROM mzmlfiles')
        return {fn: fnid for fnid, fn in cursor.fetchall()}

    def get_spectra_id(self, fn_id, retention_time=None, scan_nr=None):
        """Returns spectra id for spectra filename and retention time"""
        cursor = self.get_cursor()
        sql = 'SELECT spectra_id FROM mzml WHERE mzmlfile_id=? '
        values = [fn_id]
        if retention_time is not None:
            sql = '{0} AND retention_time=?'.format(sql)
            values.append(retention_time)
        if scan_nr is not None:
            sql = '{0} AND scan_nr=?'.format(sql)
            values.append(scan_nr)
        cursor.execute(sql, tuple(values))
        return cursor.fetchone()[0]



from app.dataformats import peptable as peptabledata
from app.dataformats import prottable as prottabledata
from app.readers import tsv as reader
from app.actions.peptable.base import evaluate_peptide
from math import log


def generate_proteins(pepfn, proteins, pepheader, scorecol, minlog,
                      higherbetter=True, protcol=False):
    """Best peptide for each protein in a table"""
    protein_peptides = {}
    if minlog:
        higherbetter = False
    if not protcol:
        protcol = peptabledata.HEADER_MASTERPROTEINS
    for psm in reader.generate_tsv_psms(pepfn, pepheader):
        p_acc = psm[protcol]
        if ';' in p_acc:
            continue
        protein_peptides = evaluate_peptide(protein_peptides, psm, p_acc,
                                            higherbetter, scorecol,
                                            fncol=False)
    if minlog:
        try:
            nextbestscore = min([pep['score'] for pep in
                                 protein_peptides.values()
                                 if pep['score'] > 0])
        except ValueError:
            import sys
            sys.stderr.write('Cannot find score of type {} which is above 0. '
                             'Only scores above zero can have a -log value. '
                             'Exiting.'.format(scorecol))
            sys.exit(1)
        nextbestscore = -log(nextbestscore, 10)
    for protein in proteins:
        try:
            peptide = protein_peptides[protein[prottabledata.HEADER_PROTEIN]]
        except KeyError:
            print('WARNING - protein {} not found in peptide '
                  'table'.format(protein[prottabledata.HEADER_PROTEIN]))
            peptide = {'score': 'NA'}
        if minlog and peptide['score'] != 'NA':
            peptide['score'] = log_score(peptide['score'], nextbestscore)
        protein[prottabledata.HEADER_QSCORE] = str(
            peptide['score'])
        yield protein


def log_score(score, nextbestscore):
    try:
        logged_score = -log(score, 10)
    except ValueError:
        logged_score = nextbestscore + 2
    return logged_score



def simple_val_fetch(feature, sqlmap, headerfields, valkey):
    val = feature[sqlmap[valkey]]
    if val is None:
        return {}
    pool = feature[sqlmap['set_name']]
    hfield = headerfields[pool]
    return {hfield: feature[sqlmap[valkey]]}


def fill_mergefeature(outfeature, iso_fun, ms1_fun, prob_fun, fdr_fun, pep_fun,
                      pdata_fun, feature, sqlfieldmap, headerfields,
                      featuredata_map, accfield=False):
    check_feat = {k: v for k, v in outfeature.items()}
    for fun in [iso_fun, ms1_fun, prob_fun, fdr_fun, pep_fun]:
        outfeature.update(fun(feature, sqlfieldmap, headerfields))
    if outfeature == check_feat:
        return
    outfeature.update(pdata_fun(outfeature, featuredata_map, headerfields, accfield))


def get_isobaric_quant(feature, sqlmap, headerfields):
    chan = feature[sqlmap['channel']]
    pool = feature[sqlmap['set_name']]
    psmfield = feature[sqlmap['isoq_psmsfield']]
    quant = feature[sqlmap['isoq_val']]
    nopsms = feature[sqlmap['isoq_psms']]
    if quant is None:
        return {}
    qfeat = {headerfields['isoquant'][chan][pool]: quant}
    if psmfield is not None:
        qfeat[headerfields['isoquant'][psmfield][pool]] = nopsms
    return qfeat



import subprocess
from app.drivers.pycolator import base
from app.actions.pycolator import qvality as preparation
from app.writers import pycolator as writers
from app.drivers.options import pycolator_options


class QvalityDriver(base.PycolatorDriver):
    """Runs qvality from two Percolator XML files. One containing target
    PSMs or peptides, and the other containing decoys."""
    outsuffix = '_qvalityout.txt'
    command = 'qvality'
    commandhelp = ('Runs qvality on an inputfile: target and decoy data. '
                   'When using separate files for target and decoy, '
                   'use --decoy to specify the decoy input file, and -f '
                   'to specify feature type (psm or peptide).')

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.score_get_fun = preparation.prepare_qvality_input
        self.target = self.fn  # for clarity
        if self.decoyfn is None:
            raise Exception('Not implemented single file qvality yet')
        self.decoy = self.decoyfn
        self.qvalityoptions = []
        if self.qoptions is not None:
            for option in self.qoptions:
                self.qvalityoptions.append(option.replace('***', '--'))

    def set_options(self):
        super().set_options()
        options = self.define_options(['decoyfn', 'featuretype', 'qoptions'],
                                      pycolator_options)
        self.options.update(options)

    def extract_from_single_target_decoy_file(self):
        """Future method to split into target and decoy file"""
        # FIXME write method, use Splittd driver
        pass

    def set_features(self):
        """Creates scorefiles for qvality's target and decoy distributions"""
        self.scores = {}
        for t_or_d, feats in zip(['target', 'decoy'], [self.target,
                                                       self.decoy]):
            self.scores[t_or_d] = {}
            self.scores[t_or_d]['scores'] = self.score_get_fun(
                feats, self.featuretype, self.prepare_percolator_output)
            self.scores[t_or_d]['fn'] = '{}_qvality_input.txt'.format(t_or_d)
            writers.write_qvality_input(self.scores[t_or_d]['scores'],
                                        self.scores[t_or_d]['fn'])

    def write(self):
        """This actually runs the qvality program from PATH."""
        outfn = self.create_outfilepath(self.fn, self.outsuffix)
        command = ['qvality']
        command.extend(self.qvalityoptions)
        command.extend([self.scores['target']['fn'], self.scores['decoy']['fn'],
                        '-o', outfn])
        subprocess.call(command)



import os
from math import log

from tests.integration import basetests


class TestAddProtData(basetests.ProttableTest):
    command = 'proteindata'
    suffix = '_proteindata.txt'

    def setUp(self):
        super().setUp()
        self.dbfile = os.path.join(self.fixdir, 'mzidtsv_db.sqlite')

    def test_addprot(self):
        self.dbfile = os.path.join(self.fixdir, 'mzidtsv_db.sqlite')
        options = ['--setname', 'S1', '--dbfile', self.dbfile]
        self.run_command(options)
        self.check_protein_data('proteincentric')

    def test_genecentric(self):
        infn = 'prottable_genes.txt'
        self.infile = [os.path.join(self.fixdir, infn)]
        self.resultfn = os.path.join(self.workdir, infn + self.suffix)
        options = ['--setname', 'S1', '--dbfile', self.dbfile,
                   '--genecentric', 'genes']
        self.run_command(options)
        self.check_protein_data('genecentric')

    def test_assoccentric(self):
        infn = 'prottable_assoc.txt'
        self.infile = os.path.join(self.fixdir, infn)
        self.resultfn = os.path.join(self.workdir, infn + self.suffix)
        options = ['--setname', 'S1', '--dbfile', self.dbfile,
                   '--genecentric', 'assoc']
        self.run_command(options)
        self.check_protein_data('assoccentric')


class TestBuild(basetests.ProttableTest):
    command = 'build'
    suffix = ''
    infilename = 'built_protein_table.txt'

    def get_std_options(self):
        cmd = [self.executable, self.command, '-d', self.workdir]
        return cmd

    def test_proteincentric(self, cutoff=False):
        self.dbfile = os.path.join(self.fixdir, 'prottable_db.sqlite')
        options = ['--fdr', '--precursor', '--isobaric', '--probability',
                   '--pep', '--dbfile', self.dbfile]
        if cutoff:
            options.extend(['--mergecutoff', str(cutoff)])
        self.run_command(options)
        sql = ('SELECT p.protein_acc, bs.set_name, pf.fdr, pep.pep, pp.quant, '
               'ppr.probability FROM proteins AS p '
               'JOIN biosets AS bs '
               'JOIN protein_fdr AS pf USING(pacc_id) '
               'JOIN protein_pep AS pep USING(pacc_id) '
               'JOIN protein_precur_quanted AS pp USING(pacc_id) '
               'JOIN protein_probability AS ppr USING(pacc_id) '
               )

        self.check_build_values(sql, ['q-value', 'PEP', 'MS1 precursor area',
                                      'Protein error probability'],
                                'Protein ID', cutoff)
        sql = ('SELECT p.protein_acc, bs.set_name, pc.channel_name, '
               'pi.quantvalue, pi.amount_psms, pf.fdr FROM proteins AS p '
               'JOIN biosets AS bs '
               'JOIN protein_iso_quanted AS pi USING(pacc_id) '
               'JOIN protquant_channels AS pc USING(channel_id) '
               'JOIN protein_fdr AS pf WHERE pf.prottable_id=pc.prottable_id '
               'AND pf.pacc_id=p.pacc_id'
               )
        self.check_built_isobaric(sql, 'Protein ID', cutoff=cutoff)
        self.check_protein_data('proteincentric')

    def test_mergecutoff(self):
        self.test_proteincentric(0.0001)

    def test_genecentric(self):
        self.dbfile = os.path.join(self.fixdir, 'prottable_gene_db.sqlite')
        options = ['--fdr', '--precursor', '--isobaric', '--probability',
                   '--pep', '--dbfile', self.dbfile, '--genecentric', 'genes']
        self.run_command(options)
        sql = ('SELECT g.gene_acc, bs.set_name, gf.fdr, pep.pep , gp.quant, '
               'gpr.probability '
               'FROM genes AS g '
               'JOIN biosets AS bs '
               'JOIN gene_fdr AS gf USING(gene_id) '
               'JOIN gene_pep AS pep USING(gene_id) '
               'JOIN gene_precur_quanted AS gp USING(gene_id) '
               'JOIN gene_probability AS gpr USING(gene_id) '
               )
        self.check_build_values(sql, ['q-value', 'PEP', 'MS1 precursor area',
                                      'Protein error probability'],
                                'Gene ID')
        sql = ('SELECT g.gene_acc, bs.set_name, pc.channel_name, '
               'pi.quantvalue, pi.amount_psms FROM genes AS g '
               'JOIN biosets AS bs '
               'JOIN gene_iso_quanted AS pi USING(gene_id) '
               'JOIN genequant_channels AS pc USING(channel_id) '
               )
        self.check_built_isobaric(sql, 'Gene ID')
        self.check_protein_data('genecentric')

    def test_nopsmnrs(self):
        """Given a lookup with NO amount psm information, output a nice gene
        table without those columns"""
        self.dbfile = os.path.join(self.fixdir, 'prottable_nopsm_db.sqlite')
        options = ['--fdr', '--precursor', '--isobaric', '--probability',
                   '--pep', '--dbfile', self.dbfile, '--genecentric', 'genes']
        self.run_command(options)
        sql = ('SELECT g.gene_acc, bs.set_name, gf.fdr, pep.pep , gp.quant, '
               'gpr.probability '
               'FROM genes AS g '
               'JOIN biosets AS bs '
               'JOIN gene_fdr AS gf USING(gene_id) '
               'JOIN gene_pep AS pep USING(gene_id) '
               'JOIN gene_precur_quanted AS gp USING(gene_id) '
               'JOIN gene_probability AS gpr USING(gene_id) '
               )
        self.check_build_values(sql, ['q-value', 'PEP', 'MS1 precursor area',
                                      'Protein error probability'],
                                'Gene ID')
        sql = ('SELECT g.gene_acc, bs.set_name, pc.channel_name, '
               'pi.quantvalue, pi.amount_psms FROM genes AS g '
               'JOIN biosets AS bs '
               'JOIN gene_iso_quanted AS pi USING(gene_id) '
               'JOIN genequant_channels AS pc USING(channel_id) '
               )
        self.check_built_isobaric(sql, 'Gene ID', check_nr_psms=False)
        self.check_protein_data('genecentric')


    def test_assoccentric(self):
        self.dbfile = os.path.join(self.fixdir, 'prottable_assoc_db.sqlite')
        options = ['--fdr', '--precursor', '--isobaric', '--probability',
                   '--pep', '--dbfile', self.dbfile, '--genecentric', 'assoc']
        self.run_command(options)
        sql = ('SELECT ai.assoc_id, bs.set_name, gf.fdr, pep.pep , gp.quant, '
               'gpr.probability '
               'FROM associated_ids AS ai '
               'JOIN biosets AS bs '
               'JOIN assoc_fdr AS gf USING(gene_id) '
               'JOIN assoc_pep AS pep USING(gene_id) '
               'JOIN assoc_precur_quanted AS gp USING(gene_id) '
               'JOIN assoc_probability AS gpr USING(gene_id) '
               )
        self.check_build_values(sql, ['q-value', 'PEP', 'MS1 precursor area',
                                      'Protein error probability'],
                                'Gene Name')
        sql = ('SELECT ai.assoc_id, bs.set_name, pc.channel_name, '
               'pi.quantvalue, pi.amount_psms '
               'FROM associated_ids AS ai '
               'JOIN biosets AS bs '
               'JOIN assoc_iso_quanted AS pi USING(gene_id) '
               'JOIN genequant_channels AS pc USING(channel_id) '
               )
        self.check_built_isobaric(sql, 'Gene Name')
        self.check_protein_data('assoccentric')


class TestAddIsoquant(basetests.ProttableTest):
    command = 'addisoquant'
    suffix = '_added_isoq.txt'
    channels = ['tmt10plex_{}'.format(x) for x in ['126', '127N', '127C',
                                                   '128N', '128C', '129N',
                                                   '129C', '130N', '130C',
                                                   '131']]
    nopsms = ['{} - # quanted PSMs'.format(ch) for ch in channels]

    def test_isoquant(self):
        isotable = os.path.join(self.fixdir, 'prottable_isoquant.txt')
        options = ['--quantfile', isotable, '--isobquantcolpattern',
                   'tmt10plex', '--qaccpattern', 'Protein ID']
        self.run_command(options)
        self.isoquant_check(isotable, 'Protein ID', self.channels,
                            self.nopsms)


class TestMS1Quant(basetests.ProttableTest):
    command = 'ms1quant'
    suffix = '_ms1q.txt'

    def setUp(self):
        super().setUp()
        self.psmfile = os.path.join(
            self.fixdir, 'mzidtsv_filtered_fr1-2_proteingrouped.txt')

    def check(self):
        top_ms1 = self.get_top_psms(self.psmfile, 'Peptide', 'MS1 area')
        top_ms1 = {prot: sum(sorted(ms1s.values(), reverse=True)[:3]) /
                   len(sorted(ms1s.values())[:3])
                   for prot, ms1s in top_ms1.items()}
        for protein in self.tsv_generator(self.resultfn):
            try:
                self.assertEqual(float(protein['MS1 precursor area']),
                                 top_ms1[protein['Protein ID']])
            except ValueError:
                self.assertNotIn(protein['Protein ID'], top_ms1)

    def test_ms1(self):
        options = ['--psmtable', self.psmfile]
        self.run_command(options)
        self.check()

    def test_with_protcol(self):
        options = ['--psmtable', self.psmfile, '--protcol', '14']
        self.run_command(options)
        self.check()


class TestProbability(basetests.ProttableTest):
    command = 'probability'
    suffix = '_protprob.txt'

    def check(self):
        top_psms = self.get_top_psms(self.pepfile, 'Peptide sequence', 'PEP',
                                     lowerbetter=True)
        for protein in self.tsv_generator(self.resultfn):
            protacc = protein['Protein ID']
            expected_prob = 1
            for prob in top_psms[protacc].values():
                expected_prob *= prob
            self.assertEqual(float(protein['Protein error probability']),
                             expected_prob)

    def setUp(self):
        super().setUp()
        self.pepfile = os.path.join(self.fixdir, 'peptable.txt')

    def test_proteincentric(self):
        options = ['--peptable', self.pepfile]
        self.run_command(options)
        self.check()

    def test_protcol(self):
        options = ['--peptable', self.pepfile, '--protcol', '15']
        self.run_command(options)
        self.check()


class TestEmpty(basetests.ProttableTest):
    command = 'emptytable'
    infilename = 'mzidtsv_filtered_fr1-2_proteingrouped.txt'
    suffix = '_prottable.txt'

    def check(self):
        expected_proteins = set()
        for psm in self.tsv_generator(self.infile[0]):
            prot = psm['Master protein(s)']
            if ';' in prot:
                continue
            expected_proteins.add(prot)
        res = [x['Protein ID'] for x in
               self.tsv_generator(self.resultfn)]
        self.assertFalse(expected_proteins.symmetric_difference(res))

    def test_proteingroup(self):
        self.run_command()
        self.check()

    def test_genecentric(self):
        options = ['--protcol', '14']
        self.run_command(options)
        self.check()


class TestBestpeptide(basetests.ProttableTest):
    command = 'bestpeptide'
    suffix = '_bestpep.tsv'

    def check(self):
        # FIXME what to do if a qvalue is at exactly 0? no log possible.
        # uglyfix here is to use PEP to avoid that scenario but the real
        # code has a nextbest qvalue for that.
        top_psms = self.get_top_psms(self.pepfile, 'Peptide sequence',
                                     'PEP', lowerbetter=True)
        qscores = {prot: -log(min([x for x in psms.values()]), 10)
                   for prot, psms in top_psms.items()}
        for protein in self.tsv_generator(self.resultfn):
            res = float(protein['Q-score best peptide'])
            self.assertEqual(qscores[protein['Protein ID']], res)

    def setUp(self):
        super().setUp()
        self.pepfile = os.path.join(self.fixdir, 'peptable.txt')

    def test_proteincentric(self):
        options = ['--logscore', '--scorecolpattern', '^PEP',
                   '--peptable', self.pepfile]
        self.run_command(options)
        self.check()

    def test_protcol(self):
        options = ['--logscore', '--scorecolpattern', '^PEP',
                   '--peptable', self.pepfile, '--protcol', '15']
        self.run_command(options)
        self.check()

    def test_protcol_pattern(self):
        options = ['--logscore', '--scorecolpattern', '^PEP',
                   '--peptable', self.pepfile, '--protcolpattern', 'Master']
        self.run_command(options)
        self.check()


class TestProtFDR(basetests.ProttableTest):
    command = 'protfdr'
    suffix = '_protfdr.txt'
    infilename = 'prottable_target.txt'

    def setUp(self):
        super().setUp()
        self.decoyfn = os.path.join(self.fixdir, 'prottable_decoy.txt')

    def test_qval(self):
        options = ['--decoyfn', self.decoyfn]
        self.run_command(options)
        expectedfn = os.path.join(self.fixdir, 'protfdr.txt')
        self.check_lines(expectedfn, self.resultfn)


class TestAssocPickFDR(basetests.ProttableTest):
    command = 'pickedfdr'
    suffix = '_pickedfdr.txt'
    infilename = 'prottable_assoc_target.txt'

    def test_result(self):
        self.infilename = 'prottable_assoc_target.txt'
        self.decoyfn = os.path.join(self.fixdir, 'prottable_assoc_decoy.txt')
        options = ['--decoyfn', self.decoyfn, '--picktype', 'result']
        self.run_command(options)
        expectedfn = os.path.join(self.fixdir, 'pickfdr_assoc.txt')
        self.check_lines(expectedfn, self.resultfn)


class TestFastaPickFDR(basetests.ProttableTest):
    command = 'pickedfdr'
    suffix = '_pickedfdr.txt'
    infilename = 'prottable_ensg_target.txt'

    def test_fasta(self):
        self.decoyfn = os.path.join(self.fixdir, 'prottable_ensg_decoy.txt')
        tfasta = os.path.join(self.fixdir, 'ensembl.fasta')
        dfasta = os.path.join(self.fixdir, 'decoy_ensembl.fasta')
        options = ['--decoyfn', self.decoyfn, '--targetfasta', tfasta,
                   '--decoyfasta', dfasta, '--picktype', 'fasta']
        self.run_command(options)
        expectedfn = os.path.join(self.fixdir, 'pickfdr_ensg.txt')
        self.check_lines(expectedfn, self.resultfn)


class TestAddFDR(basetests.ProttableTest):
    command = 'fdr'
    suffix = '_protfdr.txt'
    infilename = 'prottable_target.txt'

    def check_inbetween_fdr(self, score, fdrmap):
        def get_average(fdrmap, loscore, hiscore):
            low = [float(x) for x in fdrmap[str(loscore)]]
            hi = [float(x) for x in fdrmap[str(hiscore)]]
            return (str((low[0] + hi[0]) / 2), str((low[1] + hi[1]) / 2))

        for qv_score in sorted([float(x) for x in fdrmap.keys()]):
            if score > qv_score:
                lowerscore = str(qv_score)
            elif score < qv_score:
                return get_average(fdrmap, lowerscore, qv_score)

    def test(self):
        qvfn = os.path.join(self.fixdir, 'prottable_qvality.txt')
        options = ['-q', qvfn, '--scorecolpattern', '^Q-score']
        self.run_command(options)
        fdrmap = {float(q['Score']): (q['q-value'], q['PEP'])
                  for q in self.tsv_generator(qvfn)}
        for protein in self.tsv_generator(self.resultfn):
            score = round(float(protein['Q-score best peptide']), 5)
            try:
                self.assertEqual(protein['q-value'], fdrmap[score][0])
            except KeyError:
                fdr = self.check_inbetween_fdr(float(score), fdrmap)
                self.assertEqual(protein['q-value'], fdr[0])
                self.assertEqual(protein['PEP'], fdr[1])
            else:
                self.assertEqual(protein['PEP'], fdrmap[score][1])



from app.lookups.sqlite.base import ResultLookupInterface


class ProtPepTable(ResultLookupInterface):
    table_map = {'protein': {'fntable': 'protein_tables',
                             'feattable': 'protein_group_master',
                             'isoqtable': 'protein_iso_quanted',
                             'isochtable': 'protquant_channels',
                             'prectable': 'protein_precur_quanted',
                             'fdrtable': 'protein_fdr',
                             'peptable': 'protein_pep',
                             'probabilitytable': 'protein_probability',
                             },
                 'gene': {'fntable': 'gene_tables',
                          'feattable': 'genes',
                          'isoqtable': 'gene_iso_quanted',
                          'isochtable': 'genequant_channels',
                          'prectable': 'gene_precur_quanted',
                          'fdrtable': 'gene_fdr',
                          'peptable': 'gene_pep',
                          'probabilitytable': 'gene_probability',
                          },
                 'assoc': {'fntable': 'gene_tables',
                           'feattable': 'associated_ids',
                           'isoqtable': 'assoc_iso_quanted',
                           'isochtable': 'genequant_channels',
                           'prectable': 'assoc_precur_quanted',
                           'fdrtable': 'assoc_fdr',
                           'peptable': 'assoc_pep',
                           'probabilitytable': 'assoc_probability',
                           },
                 'peptide': {'fntable': 'peptide_tables',
                             'feattable': 'peptide_sequences',
                             'isoqtable': 'peptide_iso_quanted',
                             'isochtable': 'pepquant_channels',
                             'prectable': 'peptide_precur_quanted',
                             'fdrtable': 'peptide_fdr',
                             'peptable': 'peptide_pep',
                             }
                 }

    def get_all_poolnames(self):
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT DISTINCT set_name, set_id FROM biosets ORDER BY set_name')
        return cursor

    def store_table_files(self, tables):
        self.store_many(
            'INSERT INTO {}(set_id, filename) VALUES(?, ?)'.format(
                self.table_map[self.datatype]['fntable']),
            tables)

    def store_quant_channels(self, quantchannels, psmnrcols):
        table = self.table_map[self.datatype]['isochtable']
        if psmnrcols:
            sql = ('INSERT INTO {}({}, channel_name, amount_psms_name) VALUES'
                   '(?, ?, ?)'.format(table, self.colmap[table][1]))
        else:
            sql = ('INSERT INTO {}({}, channel_name) VALUES'
                   '(?, ?)'.format(table, self.colmap[table][1]))
        self.store_many(sql, quantchannels)

    def get_quantchannel_map(self):
        outdict = {}
        table = self.table_map[self.datatype]['isochtable']
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT channel_id, {}, channel_name, amount_psms_name'
            ' FROM {}'.format(self.colmap[table][1], table))
        for channel_id, fnid, channel_name, amount_psms_name in cursor:
            try:
                outdict[fnid][channel_name] = (channel_id, amount_psms_name)
            except KeyError:
                outdict[fnid] = {channel_name: (channel_id, amount_psms_name)}
        return outdict

    def store_isobaric_quants(self, quants, psmnrcols):
        table = self.table_map[self.datatype]['isoqtable']
        if psmnrcols:
            sql = ('INSERT INTO {}({}, channel_id, quantvalue, amount_psms) ' 
                   'VALUES ' '(?, ?, ?, ?)'.format(table, self.colmap[table][1]))
        else:
            sql = ('INSERT INTO {}({}, channel_id, quantvalue) ' 
                   'VALUES ' '(?, ?, ?)'.format(table, self.colmap[table][1]))
        self.store_many(sql, quants)

    def get_tablefn_map(self):
        table = self.table_map[self.datatype]['fntable']
        cursor = self.get_cursor()
        cursor.execute('SELECT * FROM {} '.format(table))
        return {fn: table_id for (table_id, setid, fn) in cursor}

    def get_feature_map(self):
        if self.datatype == 'protein':
            table = 'proteins'
        else:
            table = self.table_map[self.datatype]['feattable']
        columns = self.colmap[table][0:2]
        cursor = self.get_cursor()
        cursor.execute('SELECT {}, {} FROM {}'.format(columns[0], columns[1],
                                                      table))
        return {acc: table_id for (table_id, acc) in cursor}

    def store_singlecol(self, tablekey, vals):
        table = self.table_map[self.datatype][tablekey]
        cols = self.colmap[table]
        self.store_many('INSERT INTO {}({}, {}, {}) '
                        'VALUES (?, ?, ?)'.format(table, cols[0], cols[1],
                                                  cols[2]), vals)

    def store_precursor_quants(self, quants):
        self.store_singlecol('prectable', quants)

    def store_fdr(self, fdr):
        self.store_singlecol('fdrtable', fdr)

    def store_pep(self, pep):
        self.store_singlecol('peptable', pep)

    def store_probability(self, probabilities):
        self.store_singlecol('probabilitytable', probabilities)

    def update_selects(self, selectmap, fields, fieldcount):
        selectmap.update({field: i + fieldcount
                          for i, field in enumerate(fields)})
        fieldcount = max(selectmap.values()) + 1
        return selectmap, fieldcount

    def get_proteingroup_content(self):
        cursor = self.get_cursor()
        sql = ('SELECT pgm.protein_acc, pgc.protein_acc FROM '
               'protein_group_master AS pgm '
               'JOIN protein_group_content AS pgc USING(master_id)')
        return cursor.execute(sql)
    
    def get_protein_gene_symbol_for_map(self):
        fields = ['p.protein_acc', 'pep.sequence', 'pd.description',
                  'pcov.coverage', 'g.gene_acc', 'aid.assoc_id']
        sql = (
                'SELECT {} FROM {} AS p '
                'JOIN protein_psm USING(protein_acc) '
                'JOIN psms USING(psm_id) '
                'JOIN peptide_sequences AS pep USING(pep_id) '
                'LEFT OUTER JOIN prot_desc AS pd USING(protein_acc) '
                'LEFT OUTER JOIN protein_coverage '
                'AS pcov USING(protein_acc) '
                'LEFT OUTER JOIN genes AS g USING(protein_acc) '
                'LEFT OUTER JOIN associated_ids AS aid USING(protein_acc)'
                )
        sql = sql.format(','.join(fields), 'protein_group_master')
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_unique_peptide_nrs_base(self, acc, joins):
        sql = (
                'SELECT COUNT(sequence), set_name, {0} FROM ( '

                'SELECT sequence, set_name, {0} FROM ( '

                'SELECT DISTINCT pep.sequence, bs.set_name, acctable.{0} FROM '
                'peptide_sequences AS pep JOIN psms USING(pep_id) '
                'JOIN mzml USING(spectra_id) JOIN mzmlfiles USING(mzmlfile_id) '
                'JOIN biosets AS bs USING(set_id) '
                'JOIN protein_psm AS pp USING(psm_id) '
                '{1})'

                'GROUP BY sequence,set_name HAVING COUNT(sequence)=1) '

                'GROUP BY {0}, set_name')
        cursor = self.get_cursor()
        sql = sql.format(acc, joins)
        return cursor.execute(sql)
        
    def get_proteins_psms_for_map(self):
        """Gets protein-PSM combinations and other info for creating a map
        of protein data. This particular version is protein-group centric"""
        fields = ['p.protein_acc', 'sets.set_name', 'pep.sequence', 'psm.psm_id']
        firstjoin = ('psm_protein_groups', 'ppg', 'master_id')
        return self.get_proteins_psms('protein_group_master', fields, firstjoin)

    def get_unique_gene_psms(self, genetable, fields, firstjoin):
        """Uniques the results from get_proteins_psms so each PSM as defined
        by gene ID / setname / psm_id will only occur once"""
        lastgene = None
        gpsms_out, gp_ids = [], []
        for gpsm in self.get_proteins_psms(genetable, fields, firstjoin):
            if gpsm[0] != lastgene:
                for outpsm in gpsms_out:
                    yield outpsm
                lastgene = gpsm[0]
                gpsms_out, gp_ids = [], []
            gp_id = gpsm[0] + gpsm[1] + gpsm[3]
            if gp_id not in gp_ids:
                gp_ids.append(gp_id)
                gpsms_out.append(gpsm)
        for outpsm in gpsms_out:
            yield outpsm

    def get_proteins_psms(self, firsttable, fields, firstjoin):
        joins = [firstjoin]
        joins.extend([('psms', 'psm', 'psm_id'),
                      ('peptide_sequences', 'pep', 'pep_id'),
                      ('mzml', 'sp', 'spectra_id'),
                      ('mzmlfiles', 'mzfn', 'mzmlfile_id'),
                      ('biosets', 'sets', 'set_id'),
                      ])
        join_sql = ' '.join(['JOIN {} AS {} USING({})'.format(
            j[0], j[1], j[2]) for j in joins])
        sql = ('SELECT {} FROM {} '
               'AS p'.format(', '.join(fields), firsttable))
        sql = '{} {} ORDER BY {}, sets.set_name'.format(sql, join_sql,
                                                        fields[0])
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_isoquant_amountpsms_channels(self):
        cursor = self.get_cursor()
        cursor.execute(
            'SELECT DISTINCT channel_name, amount_psms_name '
            'FROM {}'.format(self.table_map[self.datatype]['isochtable']))
        return cursor

    def prepare_mergetable_sql(self, precursor=False, isobaric=False,
                               probability=False, fdr=False, pep=False):
        """Dynamically build SQL query to generate entries for the multi-set
        merged protein and peptide tables. E.g.

        SELECT g.gene_acc, pc.channel_name, pc.amount_psms_name,
               giq.quantvalue giq.amount_psms gfdr.fdr
        FROM genes AS g
        JOIN biosets AS bs
        JOIN gene_tables AS gt ON gt.set_id=bs.set_id
        JOIN genequant_channels AS pc ON pc.gene_table_id=gt.genetable_id
        JOIN gene_iso_quanted AS giq ON giq.gene_id=g.gene_id
             AND giq.channel_id=pc.channel_id
        JOIN gene_fdr AS gfdr ON gfdr.gene_id=g.gene_id
             AND gfdr.genetable_id=gt.genetable_id
        ORDER BY g.gene

        This is multi-set output because we join on biosets. The output is
        then parsed to its respective set by the action code.
        """
        featcol = self.colmap[self.table_map[self.datatype]['feattable']][1]
        selectmap, count = self.update_selects({}, ['p_acc', 'set_name'], 0)
        joins = []
        if self.datatype == 'protein':
            selects = ['pgm.{}'.format(featcol), 'bs.set_name']
            firstselect = 'pgm'
            joins.append(('proteins', 'g', ['pgm']))
        else:
            selects = ['g.{}'.format(featcol), 'bs.set_name']
            firstselect = 'g'
        if isobaric:
            selects.extend(['pc.channel_name',
                            'pc.amount_psms_name', 'giq.quantvalue',
                            'giq.amount_psms'])
            joins.extend([(self.table_map[self.datatype]['isochtable'], 'pc',
                           ['gt']),
                          (self.table_map[self.datatype]['isoqtable'], 'giq',
                           ['g', 'pc'], True),
                          ])
            fld = ['channel', 'isoq_psmsfield', 'isoq_val',
                   'isoq_psms']
            selectmap, count = self.update_selects(selectmap, fld, count)
        if precursor:
            selects.extend(['preq.quant'])
            joins.append((self.table_map[self.datatype]['prectable'], 'preq',
                          ['g', 'gt'], True))
            fld = ['preq_val']
            selectmap, count = self.update_selects(selectmap, fld, count)
        if probability:
            selects.extend(['gprob.probability'])
            joins.append((self.table_map[self.datatype]['probabilitytable'],
                          'gprob', ['g', 'gt'], True))
            fld = ['prob_val']
            selectmap, count = self.update_selects(selectmap, fld, count)
        if fdr:
            selects.extend(['gfdr.fdr'])
            joins.append((self.table_map[self.datatype]['fdrtable'], 'gfdr',
                          ['g', 'gt'], True))
            fld = ['fdr_val']
            selectmap, count = self.update_selects(selectmap, fld, count)
        if pep:
            selects.extend(['gpep.pep'])
            joins.append((self.table_map[self.datatype]['peptable'], 'gpep',
                          ['g', 'gt'], True))
            fld = ['pep_val']
            selectmap, count = self.update_selects(selectmap, fld, count)
        sql = ('SELECT {} FROM {} AS {} JOIN biosets AS bs '
               'JOIN {} AS gt ON gt.set_id=bs.set_id'.format(
                   ', '.join(selects),
                   self.table_map[self.datatype]['feattable'],
                   firstselect,
                   self.table_map[self.datatype]['fntable']))
        sql = self.get_sql_joins_mergetable(sql, joins, self.datatype)
        sql = '{} ORDER BY g.{}'.format(sql, featcol)
        return sql, selectmap

    def get_sql_joins_mergetable(self, sql, joins, pep_or_prot):
        protein_j_cols = {'g': 'pacc_id', 'gt': 'prottable_id',
                          'pgm': 'protein_acc'}
        peptide_j_cols = {'g': 'pep_id', 'gt': 'peptable_id'}
        gene_j_cols = {'g': 'gene_id', 'gt': 'genetable_id'}
        colpick = {'peptide': peptide_j_cols, 'protein': protein_j_cols,
                   'gene': gene_j_cols, 'assoc': gene_j_cols}
        join_cols = {'pc': 'channel_id'}
        join_cols.update(colpick[pep_or_prot])
        if joins:
            joinsql = ''
            for j in joins:
                joincmd = 'JOIN'
                if j[-1] is True:
                    joincmd = 'LEFT OUTER {}'.format(joincmd)
                joinmatchsql = ' AND '.join(['{0}.{1}={2}.{1}'.format(
                    j[1], join_cols[jj], jj) for jj in j[2]])
                joinsql = '{4} {0} {1} AS {2} ON {3}'.format(
                    joincmd, j[0], j[1], joinmatchsql, joinsql)
            sql = '{} {}'.format(sql, joinsql)
        return sql

    def get_merged_features(self, sql):
        return self.execute_sql(sql)



from lxml import etree
import re

from app.readers import xmlformatting as formatting
from app.readers import pycolator as reader


def create_merge_psm_map(peptides, ns):
    """Loops through peptides, stores sequences mapped to PSM ids."""
    psmmap = {}
    for peptide in peptides:
        seq = reader.get_peptide_seq(peptide, ns)
        psm_ids = reader.get_psm_ids_from_peptide(peptide, ns)
        for psm_id in psm_ids:
            try:
                psmmap[seq][psm_id.text] = 1
            except KeyError:
                psmmap[seq] = {psm_id.text: 2}
    for seq, psm_id_dict in psmmap.items():
        psmmap[seq] = [x for x in psm_id_dict]
    return psmmap


def merge_peptides(fns, ns):
    """Loops peptides from multiple files, fetches PSMs from
    sequence:PSM map, outputs correctly PSM mapped peptides"""
    peptides_to_map = reader.generate_peptides_multiple_fractions(fns, ns)
    psmmap = create_merge_psm_map(peptides_to_map, ns)
    peptides = reader.generate_peptides_multiple_fractions(fns, ns)
    for peptide in peptides:
        seq = reader.get_peptide_seq(peptide, ns)
        psm_ids = reader.get_psm_ids_from_peptide(peptide, ns)
        # remove current psm ids, repopulate with stored ones
        psm_ids.clear()
        for new_psm_id in psmmap[seq]:
            etree.SubElement(psm_ids, 'psm_id').text = new_psm_id
        yield formatting.string_and_clear(peptide, ns)


def target_decoy_generator(element_generator, decoy, ns):
    for el in element_generator:
        if el.attrib['{%s}decoy' % ns['xmlns']] == decoy:
            yield formatting.string_and_clear(el, ns)
        else:
            formatting.clear_el(el)


def split_target_decoy(elements, ns, filter_type):
    td = {'target': 'false', 'decoy': 'true'}
    feats_to_process = {'psm': None, 'peptide': None}
    for feat in feats_to_process:
        feats_to_process[feat] = target_decoy_generator(
            elements[feat], td[filter_type], ns)
    return feats_to_process


def protein_header_split_generator(elements, headers, ns):
    """Loop through proteins of each PSM/peptide. If a protein does not
    match any of headers, discard PSM/peptide immediately"""
    for el in elements:
        header_not_matching = False
        for protein in el.findall('{%s}protein_id' % ns['xmlns']):
            if not any((re.search(h, protein.text) for h in headers)):
                header_not_matching = True
                break
        if header_not_matching:
            formatting.clear_el(el)
        else:
            yield formatting.string_and_clear(el, ns)


def split_protein_header_id_type(elements, ns, protheaders):
    headers = protheaders.strip(';').split(';')
    return {x: protein_header_split_generator(elements[x], headers, ns)
            for x in ['psm', 'peptide']}



from app.readers import mzidplus as readers
from app.readers import tsv as tsvreader
from app.dataformats import mzidtsv as mzidtsvdata


def add_percolator_to_mzidtsv(mzidfn, tsvfn, multipsm, oldheader):
    """Takes a MSGF+ tsv and corresponding mzId, adds percolatordata
    to tsv lines. Generator yields the lines. Multiple PSMs per scan
    can be delivered, in which case rank is also reported.
    """
    namespace = readers.get_mzid_namespace(mzidfn)
    try:
        xmlns = '{%s}' % namespace['xmlns']
    except TypeError:
        xmlns = ''
    specfnids = readers.get_mzid_specfile_ids(mzidfn, namespace)
    mzidpepmap = {}
    for peptide in readers.generate_mzid_peptides(mzidfn, namespace):
        pep_id, seq = readers.get_mzid_peptidedata(peptide, xmlns)
        mzidpepmap[pep_id] = seq
    mzidpercomap = {}
    for specid_data in readers.generate_mzid_spec_id_items(mzidfn, namespace,
                                                           xmlns, specfnids):
        scan, fn, pepid, spec_id = specid_data
        percodata = readers.get_specidentitem_percolator_data(spec_id, xmlns)
        try:
            mzidpercomap[fn][scan][mzidpepmap[pepid]] = percodata
        except KeyError:
            try:
                mzidpercomap[fn][scan] = {mzidpepmap[pepid]: percodata}
            except KeyError:
                mzidpercomap[fn] = {scan: {mzidpepmap[pepid]: percodata}}
    for line in tsvreader.generate_tsv_psms(tsvfn, oldheader):
        outline = {k: v for k, v in line.items()}
        fn = line[mzidtsvdata.HEADER_SPECFILE]
        scan = line[mzidtsvdata.HEADER_SCANNR]
        seq = line[mzidtsvdata.HEADER_PEPTIDE]
        outline.update(mzidpercomap[fn][scan][seq])
        yield outline


def get_header_with_percolator(oldheader, multipsm=False):
    percoheader = mzidtsvdata.PERCO_HEADER
    if multipsm is True:
        # FIXME should this be here???
        percoheader.append('rank')
    ix = oldheader.index(mzidtsvdata.HEADER_EVALUE) + 1
    return oldheader[:ix] + percoheader + oldheader[ix:]



import argparse

shared_options = {
    'fn': {'driverattr': 'fn', 'dest': 'infile', 'type': 'file', 'clarg': '-i',
           'help': 'Input file of {} format'},
    'outfile': {'driverattr': 'outfile', 'dest': 'outfile', 'type': str,
                'clarg': '-o', 'help': 'Output file', 'required': False},
    'outdir': {'driverattr': 'outdir', 'dest': 'outdir', 'clarg': '-d',
               'help': 'Directory to output in', 'type': 'file',
               'required': False},
    'multifiles': {'driverattr': 'fn', 'dest': 'infile', 'clarg': '-i',
                   'help': 'Multiple input files of {} format',
                   'type': 'file', 'nargs': '+'},
    'lookupfn': {'driverattr': 'lookupfn', 'clarg': '--dbfile',
                 'type': 'file', 'help': 'Database lookup file'},
    'setnames': {'driverattr': 'setnames', 'dest': 'setnames',
                 'type': str, 'nargs': '+', 'clarg': '--setnames',
                 'help': 'Names of biological sets. Can be '
                 'specified with quotation marks if spaces are '
                 'used'},
    'spectracol': {'driverattr': 'spectracol', 'dest': 'spectracol',
                   'type': int, 'clarg': '--spectracol', 'help':
                   'Column number in which spectra file names are, '
                   'in case some framework has changed the file '
                   'names. First column number is 1.', 'required': False,
                   'default': 1},
    'decoyfn': {'driverattr': 'decoyfn', 'dest': 'decoyfn',
                'help': 'Decoy input file (percolator out XML) for qvality',
                'type': 'file', 'clarg': '--decoyfn'},
    'proline': {'driverattr': 'proline', 'dest': 'proline', 'required': False,
                'clarg': '--cutproline', 'action': 'store_const',
                'const': True, 'default': False, 'help': 'Flag to make '
                'trypsin before a proline residue. Then filtering will be '
                'done against both cut and non-cut peptides.',
                },
    'fasta': {'driverattr': 'fasta', 'dest': 'fasta',
              'type': 'file', 'help': 'FASTA sequence database',
              'required': False, 'default': False, 'clarg': '--fasta'},
    'trypsinize': {'driverattr': 'trypsinize', 'dest': 'trypsinize',
                   'clarg': '--notrypsin', 'required': False,
                   'action': 'store_const', 'const': False, 'default': True,
                   'help': 'Do not trypsinize. User is expected to deliver a'
                   'pretrypsinized FASTA file'
                   },
    'featuretype': {'driverattr': 'featuretype', 'dest': 'featuretype',
                    'help': 'Feature type to use for qvality. Can either be '
                    'psm or peptide.', 'clarg': '--feattype',
                    'type': 'pick', 'picks': ['psm', 'peptide']},
    'unroll': {'driverattr': 'unroll', 'clarg': '--unroll', 'const': True,
               'action': 'store_const', 'default': False, 'help': 'PSM table '
               'from Mzid2TSV contains either one PSM per line with all '
               'the proteins of that shared peptide on the same line (not'
               ' unrolled, default), or one PSM/protein match per line '
               'where each protein from that shared peptide gets its own '
               'line (unrolled).', 'required': False},
    'genecentric': {'driverattr': 'genecentric', 'dest': 'genecentric',
                    'clarg': '--genecentric', 'type': 'pick',
                    'picks': ['assoc', 'genes'], 'required': False,
                    'default': False, 'help': 'Do not include protein group '
                    'data in output. Should be one of [genes, assoc]. '
                    'With assoc, associated gene IDs are used from e.g. '
                    'Biomart rather than the ones found in the FASTA db used '
                    'for PSM search. These need to have been stored when '
                    'creating a PSM lookup.'},
    'isobaric': {'driverattr': 'isobaric', 'clarg': '--isobaric',
                 'action': 'store_const', 'const': True, 'default': False,
                 'help': 'Specifies to add isobaric quant data from lookup DB '
                 'to output table', 'required': False,
                 },
    'precursor': {'driverattr': 'precursor', 'clarg': '--precursor',
                  'action': 'store_const', 'const': True, 'default': False,
                  'help': 'Specifies to add precursor quant data from lookup '
                  'DB to output table', 'required': False,
                  },
    'quantcolpattern': {'driverattr': 'quantcolpattern',
                        'clarg': '--isobquantcolpattern', 'type': str,
                        'default': None, 'required': False,
                        'help': 'Unique text pattern to identify '
                        'isobaric quant columns in input table.'},
    'precursorquantcolpattern': {'driverattr': 'precursorquantcolpattern',
                                 'type': str, 'required': False,
                                 'dest': 'precursorquantcolpattern',
                                 'clarg': '--ms1quantcolpattern',
                                 'default': None,
                                 'help': 'Unique text pattern to identify '
                                 'precursor quant column in input table.'},
    'quantacccolpattern': {'driverattr': 'quantacccolpattern',
                           'clarg': '--qaccpattern', 'type': str,
                           'help': 'Unique text pattern to identify '
                           'accession column in table containing quant info.'},
    'qvalityout': {'driverattr': 'qvalityout', 'dest': 'qvalityout',
                   'help': 'Qvality output file to fetch q-values and PEP '
                   'from', 'type': 'file', 'clarg': ['-q', '--qvality']},
    'proteincol': {'driverattr': 'proteincol', 'clarg': '--protcol',
                   'type': int, 'required': False, 'help': 'Column number in '
                   'table in which protein or gene accessions are. '
                   'stored. First column number is 1. Use in case of not '
                   'using standard {} column'},
    'pcolpattern': {'driverattr': 'pcolpattern', 'clarg': '--protcolpattern',
                    'type': str, 'required': False, 'help': 'Text pattern to '
                    'identify column in table in which protein or gene '
                    'accessions are. Use in case of not using standard '
                    '{} column', 'default': False},
    'fdrcolpattern': {'driverattr': 'fdrcolpattern', 'dest': 'fdrcolpattern',
                      'clarg': '--fdrcolpattern', 'type': str,
                      'required': False, 'default': None,
                      'help': 'Unique text pattern to identify '
                      'FDR column in input table.'},
    'fastadelim': {'driverattr': 'fastadelim', 'clarg': '--fastadelim',
                   'dest': 'fastadelim', 'required': False, 'type': 'pick',
                   'picks': ['tab', 'pipe', 'semicolon'],
                   'help': 'Delimiter in FASTA header, used to parse gene '
                   'names in case of non-ENSEMBL/Uniprot'},
    'genefield': {'driverattr': 'genefield', 'clarg': '--genefield',
                  'dest': 'genefield', 'required': False, 'type': int,
                  'help': 'Field nr (first=1) in FASTA that contains gene '
                  'name when using --fastadelim to parse the gene names'},
    'minlength': {'driverattr': 'minlength', 'dest': 'minlength', 'default': 0,
                  'help': 'Minimum length of peptide to be included',
                  'type': int, 'clarg': '--minlen', 'required': False},
}

mslookup_options = {
    'falloff': {'driverattr': 'falloff', 'dest': 'falloff',
                'clarg': '--insourcefrag', 'default': False,
                'action': 'store_const', 'const': True, 'help': 'Apply '
                'filter against both intact peptides and those '
                'that match to the C-terminal part of a tryptic peptide '
                'from the database, resulting from in-source fragmentation, '
                'where some amino acids will be missing from the N-terminus. '
                'Specify the max number of amino acids that may be missing. '
                'Database should be built with this '
                'flag in order for the lookup to work, since sequences '
                'will be stored and looked up reversed', 'required': False
                },
    'mapfn': {'driverattr': 'mapfn', 'dest': 'mapfn',
              'type': 'file', 'clarg': '--map',
              'required': False, 'help': 'File that contains '
              'a map obtained from ENSEMBL BioMart which '
              'should contain mappings from protein accession '
              'to Gene ENSG and Symbol.'},
    'decoy': {'driverattr': 'decoy', 'dest': 'decoy', 'clarg': '--decoy',
              'action': 'store_const', 'const': True,
              'default': False, 'help': 'Specifies lookup is '
              'for decoy PSMs, use with --map in case there '
              'are no decoy symbols in the FASTA used to '
              'search.', 'required': False},
    'spectrafns': {'driverattr': 'spectrafns', 'dest': 'spectra',
                   'type': str, 'help': 'Spectra files in mzML '
                   'format. Multiple files can be specified, if '
                   'order is important, e.g. when matching them '
                   'with quant data, the order will be their input '
                   'order at the command line.', 'clarg': '--spectra',
                   'nargs': '+'},
    'quantfiletype': {'driverattr': 'quantfiletype', 'dest': 'quanttype',
                      'clarg': '--quanttype', 'type': 'pick', 'help':
                      'Filetype of '
                      'precursor quants to store. One of kronik or openms.',
                      'picks': ['kronik', 'openms']},
    'rttol': {'driverattr': 'rt_tol', 'dest': 'rttol', 'clarg': '--rttol',
              'type': float, 'help': 'Specifies tolerance in seconds for '
              'retention time when mapping MS1 feature quant info to '
              'identifications in the PSM table.'},
    'mztol': {'driverattr': 'mz_tol', 'dest': 'mztol', 'clarg': '--mztol',
              'type': float, 'help': 'Specifies tolerance in mass-to-charge '
              'when mapping MS1 feature quant info to identifications in '
              'the PSM table.'},
    'mztoltype': {'driverattr': 'mz_toltype', 'dest': 'mztoltype',
                  'type': 'pick', 'picks': ['ppm', 'Da'],
                  'clarg': '--mztoltype',
                  'help': 'Type of tolerance in mass-to-charge when mapping '
                  'MS1 feature quant info to identifications in the PSM table.'
                  ' One of ppm, Da.'},
    'peptidecol': {'driverattr': 'peptidecol', 'dest': 'peptidecol',
                   'type': int, 'clarg': '--peptidecol', 'help':
                   'Column nr of peptide table where peptide sequences are '
                   'stored. First column is nr. 1'},
    'psmnrcolpattern': {'driverattr': 'psmnrcolpattern',
                        'dest': 'psmnrcolpattern',
                        'clarg': '--psmnrcolpattern',
                        'type': str, 'default': None, 'required': False,
                        'help': 'Unique text pattern to identify '
                        'number-of-psms column in input table.'},
    'probcolpattern': {'driverattr': 'probcolpattern',
                       'dest': 'probcolpattern', 'clarg': '--probcolpattern',
                       'type': str, 'required': False,
                       'default': None,
                       'help': 'Unique text pattern to identify '
                       'protein probability column in input table.'},
    'pepcolpattern': {'driverattr': 'pepcolpattern', 'dest': 'pepcolpattern',
                      'clarg': '--pepcolpattern', 'type': str,
                      'required': False, 'default': None,
                      'help': 'Unique text pattern to identify '
                      'protein PEP column in input table.'},
}
mslookup_options['proteincol'] = {k: v for k, v
                                  in shared_options['proteincol'].items()}
mslookup_options['proteincol'].update(
    {'default': 1, 'help':
     mslookup_options['proteincol']['help'].format('first')})
mslookup_options['fasta'] = {k: v for k, v in shared_options['fasta'].items()}
mslookup_options['fasta']['help'] = ('FASTA sequence database to use when '
                                     'extracting gene names to the PSM '
                                     'table from proteins')

pycolator_options = {
    'maxlength': {'driverattr': 'maxlength', 'dest': 'maxlength',
                  'default': None, 'required': False,
                  'help': 'Maximum length of peptide to be included in '
                  'filtered data.', 'type': int, 'clarg': '--maxlen'},
    'score': {'driverattr': 'score', 'dest': 'score', 'default': 'svm',
              'help': 'Score to filter unique peptides on, e.g. svm',
              'type': 'pick', 'picks': ['svm', 'q', 'pep', 'p'],
              'clarg': '--score', 'required': False},
    'qoptions': {'driverattr': 'qoptions', 'dest': 'qoptions',
                 'required': False, 'clarg': '--qoptions', 'default': None,
                 'help': 'Extra options that may be passed to qvality. '
                 'Option form: --qoptions ***flag value ***flag ***flag value',
                 'nargs': '+', 'type': str},
    'protheaders': {'driverattr': 'protheaders', 'clarg': '--protheaders',
                    'nargs': '+', 'type': str,
                    'help': 'Specify protein FASTA headers to split on. '
                    'Multiple headers of the same split-type can be grouped '
                    'with semicolons. E.g. --protheaders ENSP;sp '
                    'PSEUDOGEN;ncRNA would split into ENSP/swissprot peptides '
                    'and pseudogenes/non-coding RNA peptides.'},
    'deamidate': {'driverattr': 'deamidate', 'clarg': '--deamidate',
                  'action': 'store_const', 'default': False, 'const': True,
                  'help': 'Filter against both normal peptides and deamidated '
                  'peptides where a D->N transition has occurred.',
                  'required': False},
    'forcetryp': {'driverattr': 'forcetryp', 'clarg': '--enforce-tryptic',
                  'action': 'store_const', 'default': False, 'const': True,
                  'help': 'When filtering peptides against whole proteins, '
                  'filter out peptides that match a whole protein but only '
                  'if they are fully tryptic, i.e. the protein needs K,R 1 '
                  'position upstream of the peptide, and the peptide '
                  'C-terminal should also be K,R. Useful when discerning '
                  'from pseudogenes', 'required': False},
    'falloff': {'driverattr': 'falloff', 'dest': 'falloff',
                'clarg': '--insourcefrag',
                'type': int, 'default': 0, 'help': 'Apply '
                'filter against both intact peptides and those '
                'that match to the C-terminal part of a tryptic peptide '
                'from the database, resulting from in-source fragmentation, '
                'where some amino acids will be missing from the N-terminus. '
                'Specify the max number of amino acids that may be missing. '
                'Database should be built with this '
                'flag in order for the lookup to work, since sequences '
                'will be stored and looked up reversed', 'required': False
                },
}

mzidtsv_options = {
    'confcol': {'driverattr': 'confcol', 'clarg': '--confidence-col',
                'help': 'Confidence column number or name in the tsv file. '
                'First column has number 1.', 'type': int, 'required': False},
    'confpattern': {'driverattr': 'confpattern', 'clarg': '--confcolpattern',
                    'type': str, 'required': False, 'help': 'Text pattern to '
                    'identify column in table on which confidence filtering '
                    'should be done. Use in case of not using standard '
                    '{} column', 'default': False},
    'conflvl': {'driverattr': 'conflvl', 'clarg': '--confidence-lvl',
                'help': 'Confidence cutoff level as a floating point number',
                'type': float},
    'conftype': {'driverattr': 'conftype', 'clarg': '--confidence-better',
                 'help': 'Confidence type to define if higher or lower score '
                 'is better. One of [higher, lower]', 'type': 'pick',
                 'picks': ['higher', 'lower']},
    'medianpsms': {'driverattr': 'medianpsms', 'clarg': '--medianpsms',
                   'help': 'In case of using a separate PSM table with more '
                   'data to generate more robust medians (i.e. a superset of '
                   'input PSMs), specify that file here. The normalization '
                   'factors will be calculated on this file, and the PSMs '
                   'in the input will be adjusted using those factors rather '
                   'than factors derived from solely their own quantification '
                   'data', 'type': 'file', 'required': False},
    'normalize': {'driverattr': 'normalize', 'clarg': '--normalize',
                  'type': str, 'default': False, 'required': False,
                  'help': 'Normalization method for isobaric '
                  'quant data on protein or peptide level. Currently only '
                  'median centering is used. Use "--normalize median"'},
    'normalizeratios': {'driverattr': 'normalizeratios', 'type': 'file',
                        'clarg': '--norm-ratios', 'required': False,
                        'default': False, 'help': 'In case of using a '
                        'separate table to generate channel medians for '
                        'normalizing, specify that file here. The '
                        'normalization factors will be calculated from this '
                        'file, and the features in the input will be adjusted '
                        'using those factors rather than factors derived from '
                        'their own quantification data'},
    'targettable': {'driverattr': 'targettable', 'clarg': '--targettable',
                    'help': 'Table to output PSM or other feature quant data '
                    'to. Used when calculating PSM isobaric intenstity ratios '
                    'for proteins, peptides, genes. Leaving empty will output '
                    'to a new table, or when no --protcol is specified, '
                    'pastes ratios to the PSM table they are fetched from.',
                    'type': 'file', 'required': False},
    'mzidfn': {'driverattr': 'mzidfn', 'clarg': '--mzid', 'help': 'mzIdentML',
               'type': 'file'},
    'bioset': {'driverattr': 'bioset', 'clarg': '--bioset', 'const': True,
               'action': 'store_const', 'default': False,
               'help': 'this enables automatic splitting on '
               'biological set names, for which a a column specifying '
               'these must exist.', 'required': False},
    'splitcol': {'driverattr': 'splitcol', 'clarg': '--splitcol', 'type': int,
                 'help': 'Column number to split a PSM table on. First column '
                 'is number 1', 'required': False, 'default': None,
                 },
    'denompatterns': {'driverattr': 'denompatterns', 'required': False,
                      'clarg': '--denompatterns', 'type': str, 'nargs': '+',
                      'help': 'Regex patterns to detect denominator channels '
                      'when creating a PSM table with normalized ratios. If '
                      'both patterns and column numbers are given then column '
                      'numbers are used. Usage e.g. --denompattern _126 _131. '
                      'Also possible: --denompattern _12[6-7] to detect '
                      'multiple columns.'
                      },
    'denomcols': {'driverattr': 'denomcols', 'clarg': '--denomcols',
                  'type': int, 'nargs': '+', 'required': False,
                  'help': 'Column numbers of denominator channels when '
                  'creating a PSM table with normalized ratios',
                  },
    'minint': {'driverattr': 'minint', 'clarg': '--minint', 'type': float,
               'help': 'Intensity threshold of PSMs when calculating '
               'isobaric ratios. Values below threshold will be set to NA.',
               'required': False, 'default': -1,
               },
}
mzidtsv_options['quantcolpattern'] = {
    k: v for k, v in shared_options['quantcolpattern'].items()}
mzidtsv_options['quantcolpattern']['required'] = True

pepprottable_options = {
    # a mock infile to make sure we don't show or need an infile, e.g. in
    # case of building something from lookup
    'mock_infn': {'driverattr': 'fn', 'clarg': '-i', 'required': False,
                  'help': argparse.SUPPRESS, 'type': 'file'},
    'fdr': {'driverattr': 'fdr', 'clarg': '--fdr', 'action': 'store_const',
            'default': False, 'const': True, 'required': False,
            'help': 'Output FDR data to table'},
    'pep': {'driverattr': 'pep', 'clarg': '--pep', 'action': 'store_const',
            'default': False, 'const': True, 'required': False,
            'help': 'Output posterior error probabilities (PEP) to table.'},
    'quantfile': {'driverattr': 'quantfile', 'clarg': '--quantfile',
                  'type': 'file', 'help': 'File containing isobaric quant '
                  'data to add to table.'},
    'scorecolpattern': {'driverattr': 'scorecolpattern', 'type': str,
                        'clarg': '--scorecolpattern', 'help': 'Regular '
                        'expression pattern to find column where score '
                        'to use (e.g. percolator svm-score) is written.'},

}


prottable_options = {k: v for k, v in pepprottable_options.items()}
prottable_options['proteincol'] = {k: v for k, v
                                   in shared_options['proteincol'].items()}
prottable_options['proteincol'].update(
    {'default': False, 'help':
     prottable_options['proteincol']['help'].format('Master protein')})
prottable_options.update({
    'setname': {'driverattr': 'setname', 'clarg': '--setname', 'type': str,
                'help': 'Name of biological set to use when adding protein '
                'info to the table'},
    'probability': {'driverattr': 'probability', 'clarg': '--probability',
                    'action': 'store_const', 'default': False, 'const': True,
                    'help': 'Output protein probability to table.',
                    'required': False},
    'psmfile': {'driverattr': 'psmfile', 'clarg': '--psmtable', 'type': 'file',
                'help': 'PSM table file containing precursor quant data to '
                'add to table.'},
    'pepfile': {'driverattr': 'pepfile', 'clarg': '--peptable', 'type': 'file',
                'help': 'Peptide table file'},
    'minlogscore': {'driverattr': 'minlogscore', 'clarg': '--logscore',
                    'action': 'store_const', 'default': False, 'const': True,
                    'required': False, 'help': 'Score, e.g. q-values will '
                    'be converted to -log10 values.'},
    't_fasta': {'driverattr': 't_fasta', 'clarg': '--targetfasta',
                'type': 'file', 'help': 'FASTA file with target proteins '
                'to determine best scoring proteins of target/decoy pairs '
                'for pickqvality. In case using --picktype fasta',
                'required': False},
    'd_fasta': {'driverattr': 'd_fasta', 'clarg': '--decoyfasta',
                'type': 'file', 'help': 'FASTA file with decoy proteins '
                'to determine best scoring proteins of target/decoy pairs '
                'for pickqvality. In case using --picktype fasta',
                'required': False},
    'picktype': {'driverattr': 'picktype', 'clarg': '--picktype',
                 'type': 'pick', 'picks': ['fasta', 'result'],
                 'help': 'Feature type to use for qvality. Can be one of '
                 '[fasta, result].'},
    'mergecutoff': {'driverattr': 'mergecutoff', 'clarg': '--mergecutoff',
                    'type': float, 'default': False, 'help': 'FDR cutoff when '
                    'building merged protein table, to use when a cutoff has '
                    'been used before storing the table to lookup. FDR values '
                    'need to be stored in the lookup', 'required': False},
})


peptable_options = {k: v for k, v in pepprottable_options.items()}
peptable_options.update({
    'spectracol': {'driverattr': 'spectracol', 'dest': 'spectracol',
                   'type': int, 'clarg': '--spectracol', 'help':
                   'Specify this column number (first col. is 1) '
                   'containing PSM table spectrafiles (e.g. mzML) '
                   'if you want to track PSMs when creating peptide '
                   'tables', 'required': False},
    'noncentric': {'driverattr': 'noncentric', 'dest': 'noncentric',
                   'clarg': '--noncentric', 'action': 'store_const',
                   'const': True, 'default': False, 'required': False,
                   'help': 'Do not include protein group '
                   'or gene data in output, just use accessions. '},
    'genecentric': {'driverattr': 'genecentric', 'dest': 'genecentric',
                    'clarg': '--genecentric', 'action': 'store_const',
                    'const': True, 'default': False, 'required': False,
                    'help': 'Do not include protein group '
                    'data in output, but use gene names instead. '
                    'These need to have been stored when '
                    'creating a PSM lookup.'},
})



from collections import OrderedDict

from app.readers import fasta as fastareader
from app.readers import tsv as tsvreader
from app.dataformats import mzidtsv as mzidtsvdata


def create_header(oldheader):
    p_ix = oldheader.index(mzidtsvdata.HEADER_PROTEIN) + 1
    newfields = [mzidtsvdata.HEADER_GENE, mzidtsvdata.HEADER_SYMBOL,
                 mzidtsvdata.HEADER_DESCRIPTION]
    return oldheader[:p_ix] + newfields + oldheader[p_ix:]


def add_genes_to_psm_table(psmfn, oldheader, pgdb):
    gpmap = pgdb.get_protein_gene_map()
    for psm in tsvreader.generate_tsv_psms(psmfn, oldheader):
        outpsm = {x: y for x, y in psm.items()}
        proteins = tsvreader.get_proteins_from_psm(psm)
        outpsm[mzidtsvdata.HEADER_GENE] = ';'.join(get_genes(proteins, gpmap))
        symbols = get_symbols(proteins, gpmap)
        desc = get_descriptions(proteins, gpmap)
        outpsm[mzidtsvdata.HEADER_SYMBOL] = ';'.join(symbols)
        outpsm[mzidtsvdata.HEADER_DESCRIPTION] = ';'.join(desc)
        yield outpsm


def get_mapped(proteins, gpmap, outtype):
    # FIXME multiple vals for a protein?
    outvals = [gpmap[protein][outtype] for protein in proteins]
    outvals = OrderedDict([(val, 1) for val in outvals]).keys()
    if None in outvals:
        return ['NA']
    else:
        return outvals


def get_genes(proteins, gpmap):
    return get_mapped(proteins, gpmap, 'gene')


def get_descriptions(proteins, gpmap):
    descriptions = get_mapped(proteins, gpmap, 'desc')
    return [x.replace('\t', ' ') for x in descriptions]


def get_symbols(proteins, gpmap):
    return get_mapped(proteins, gpmap, 'symbol')



from app.drivers.prottable.base import ProttableAddData
from app.actions.prottable import probability as preparation
from app.readers import tsv as reader
from app.drivers.options import prottable_options


class AddProteinProbability(ProttableAddData):
    outsuffix = '_protprob.txt'
    command = 'probability'
    commandhelp = ('Add protein probabilities from peptide table posterior '
                   'error probabilities (PEP). Probabilities are calculated '
                   'as in Nesvizhskii et al. (2003) Anal.Chem., eq 3.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['pepfile', 'proteincol'],
                                                prottable_options))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.headertypes = ['probability']

    def initialize_input(self):
        super().initialize_input()
        self.in_peptides = reader.generate_tsv_peptides(self.pepfile)
        pepheader = reader.get_tsv_header(self.pepfile)
        self.get_column_header_for_number(['proteincol'], pepheader)

    def set_feature_generator(self):
        self.features = preparation.add_nesvi_protein_probability(
            self.in_proteins, self.in_peptides, self.headerfields,
            self.proteincol)



from app.drivers.prottable.base import PepProttableDriver
from app.actions.headers import peptable as head
from app.readers import tsv as tsvreader
from app.actions.peptable import model_qvals as prep
from app.drivers.options import peptable_options


class ModelQValuesDriver(PepProttableDriver):
    """Given a peptide table, this uses linear regression to model the
    peptide q-values against a score, e.g. svm-score.
    # FIXME
    It currently also removes the column with PEPs, since it will no
    longer be correct.
    """
    outsuffix = '_qmodel.txt'
    command = 'modelqvals'
    commandhelp = ('Recalculate peptide q-values by creating a linear model '
                   'of them against a score (partial least squares '
                   'regression).')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['scorecolpattern',
                                                 'fdrcolpattern'],
                                                peptable_options))
        self.options['--fdrcolpattern'].update({'required': True})

    def initialize_input(self):
        self.oldheader = tsvreader.get_tsv_header(self.fn)
        self.scorecol = tsvreader.get_cols_in_file(self.scorecolpattern,
                                                   self.oldheader, True)
        self.qvalcol = tsvreader.get_cols_in_file(self.fdrcolpattern,
                                                  self.oldheader, True)

    def create_header(self):
        self.header = head.get_linear_model_header(self.oldheader)

    def set_feature_generator(self):
        self.features = prep.recalculate_qvals_linear_model(self.fn,
                                                            self.scorecol,
                                                            self.qvalcol,
                                                            10e-4)



from app.readers import tsv as readers
from app.dataformats import mzidtsv as mzidtsvdata


def generate_psms_quanted(quantdb, tsvfn, isob_header, oldheader,
                          isobaric=False, precursor=False):
    """Takes dbfn and connects, gets quants for each line in tsvfn, sorts
    them in line by using keys in quantheader list."""
    allquants, sqlfields = quantdb.select_all_psm_quants(isobaric, precursor)
    quant = next(allquants)
    for rownr, psm in enumerate(readers.generate_tsv_psms(tsvfn, oldheader)):
        outpsm = {x: y for x, y in psm.items()}
        if precursor:
            pquant = quant[sqlfields['precursor']]
            if pquant is None:
                pquant = 'NA'
            outpsm.update({mzidtsvdata.HEADER_PRECURSOR_QUANT: str(pquant)})
        if isobaric:
            isoquants = {}
            while quant[0] == rownr:
                isoquants.update({quant[sqlfields['isochan']]:
                                  str(quant[sqlfields['isoquant']])})
                try:
                    quant = next(allquants)
                except StopIteration:
                    # last PSM, break from while loop or it is not yielded at all
                    break
            outpsm.update(get_quant_NAs(isoquants, isob_header))
        else:
            try:
                quant = next(allquants)
            except StopIteration:
                # last PSM, needs explicit yield/break or it will not be yielded
                yield outpsm
                break
        yield outpsm


def get_full_and_isobaric_headers(oldheader, quantdb, isobaric=False,
                                  precursor=False):
    # FIXME:
    # if we let db decide, and already isobaric done on tsv gets a double db,
    # then automatically we get twice isobaric on the tsv header. not good,
    # because it will fuck up the psm dicts per line.
    # also, maybe we should make sure there is no overwriting the header fields
    # with new fields in case that would  happne, or output a set.
    # is there any other scenario where we dont want a specific part of quant
    # data included in the tsv except 'it is already there'?
    # is we're outputting a set, we should do this as a general method for tsv
    # driven stuff. then output here oldheader and new fields as tuple
    # FIXME Make sure header is in mass-order, not alphabetical as it is now
    fullheader = oldheader
    if precursor:
        fullheader += [mzidtsvdata.HEADER_PRECURSOR_QUANT]
    if isobaric:
        isob_header = [x[0] for x in quantdb.get_all_quantmaps()]
        fullheader += isob_header
    else:
        isob_header = None
    return fullheader, isob_header


def get_quant_NAs(quantdata, quantheader):
    """Takes quantdata in a dict and header with quantkeys
    (eg iTRAQ isotopes). Returns dict of quant intensities
    with missing keys set to NA."""
    out = {}
    for qkey in quantheader:
        out[qkey] = quantdata.get(qkey, 'NA')
    return out



from lxml import etree

def string_and_clear(el, ns):
    str_el = stringify_strip_namespace_declaration(el, ns)
    clear_el(el)
    return str_el

def stringify_strip_namespace_declaration(el, ns):
    strxml = etree.tostring(el, encoding='utf-8').decode('utf-8')
    strxml = strxml.replace('xmlns="{0}" '.format(ns['xmlns']), '')
    strxml = strxml.replace('xmlns:p="{0}" '.format(ns['xmlns:p']), '')
    strxml = strxml.replace('xmlns:xsi="{0}" '.format(ns['xmlns:xsi']), '')
    return strxml

def clear_el(el):
    el.clear()
    while el.getprevious() is not None:
        del(el.getparent()[0])




from numpy import transpose
from collections import OrderedDict
from sqlite3 import OperationalError

from app.dataformats.prottable import HEADER_NO_PSMS_SUFFIX

def generate_general_header(headerfields, fieldtypes, firstfield, oldheader,
                            group_by_field):
    """From headerfield object, this generates a full header as a list, ready
    to write to a TSV file
    E.g:
    headerfield = {precusroquant: {HEADER_AREA: OD([(set1, set1_HEAD), (set2,
    set2_HEAD), etc])}}"""
    if not oldheader:
        header = [firstfield]
    else:
        header = [firstfield] + oldheader[1:]
    poolfields = OrderedDict()
    poolfields[None] = []  # Have non-pool/set columns come before pool-columns
    if group_by_field:
        header.extend(poolfields[None])
    for fieldtype in fieldtypes:
        try:
            fields = headerfields[fieldtype]
        except KeyError:
            continue
        if type(fields) == list:
            header.extend(fields)
        elif group_by_field:
            pfmatrix = [list(x.values()) for k, x in fields.items()
                        if not HEADER_NO_PSMS_SUFFIX in k]
            header.extend([x for y in transpose(pfmatrix) for x in y])
            if fieldtype == 'isoquant':
                pfmatrix = [list(x.values()) for k, x in fields.items()
                            if HEADER_NO_PSMS_SUFFIX in k]
                header.extend([x for y in transpose(pfmatrix) for x in y])
        else:
            for pool_field in fields.values():
                for pool, field in pool_field.items():
                    try:
                        poolfields[pool].append(field)
                    except KeyError:
                        poolfields[pool] = [field]
    if poolfields and not group_by_field:
        for fields in poolfields.values():
            header.extend(fields)
    return header


def generate_headerfields(headertypes, allfield_defs, poolnames, db=False, genecentric=False):
    """Returns a headerfields object (dict) which contains information on
    fields of the header, including optional pool names"""
    hfields = {}
    for fieldtype in headertypes:
        if fieldtype == 'isoquant':
            continue
        elif fieldtype == 'proteindata':
            hfield_definitions = allfield_defs[fieldtype](poolnames, genecentric)
        else:
            hfield_definitions = allfield_defs[fieldtype](poolnames)
        hfields[fieldtype] = OrderedDict()
        for fieldname, poolnames in hfield_definitions.items():
            hfields[fieldtype][fieldname] = get_header_field(fieldname,
                                                             poolnames)
    if 'isoquant' in headertypes:
        hfield_definitions = allfield_defs['isoquant'](db, poolnames)
        hfields['isoquant'] = OrderedDict()
        for poolname in poolnames:
            for fieldname in hfield_definitions:
                hfields['isoquant'][fieldname] = get_header_field(
                    fieldname, poolnames)
    return hfields


def get_header_field(field, poolnames=False):
    if poolnames:
        return OrderedDict([(pool, '{}_{}'.format(pool, field))
                            for pool in poolnames])
    else:
        return {None: field}


def get_isoquant_fields(pqdb=False, poolnames=False):
    """Returns a headerfield dict for isobaric quant channels. Channels are
    taken from DB and there isn't a pool-independent version of this yet"""
    # FIXME when is a None database passed?
    if pqdb is None:
        return {}
    try:
        channels_psms = pqdb.get_isoquant_amountpsms_channels()
    except OperationalError:
        # FIXME what does this catch?
        return {}
    quantheader, psmsheader = OrderedDict(), OrderedDict()
    for chan_name, amnt_psms_name in channels_psms:
        quantheader[chan_name] = poolnames
        if amnt_psms_name:
            psmsheader[amnt_psms_name] = poolnames
    quantheader.update(psmsheader)
    return quantheader



from app.actions.mzidtsv import spectra as actions
from app.drivers.mzidtsv import MzidTSVDriver
from app.drivers.options import mzidtsv_options


class TSVSpectraDriver(MzidTSVDriver):
    lookuptype = 'spectra'
    outsuffix = '_spectradata.tsv'
    command = 'specdata'
    commandhelp = ('Add spectra data such as retention time '
                   'biological set name to PSM table.')

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['lookupfn', 'spectracol'],
                                                mzidtsv_options))

    def get_psms(self):
        """Creates iterator to write to new tsv. Contains input tsv
        lines plus quant data for these."""
        self.header = actions.create_header(self.oldheader, self.spectracol)
        self.psms = actions.generate_psms_spectradata(self.lookup, self.fn,
                                                      self.oldheader)



from lxml import etree
import itertools
from app.readers import xmlformatting as formatting


def get_namespace_from_top(fn, key='xmlns'):
    ac, el = next(etree.iterparse(fn))
    return {'xmlns': el.nsmap[key]}


def get_root_el(fn):
    rootgen = etree.iterparse(fn, events=('start',))
    root = next(rootgen)[1]
    for child in root.getchildren():
        root.remove(child)
    return root


def get_namespace(fn):
    root = get_root_el(fn)
    ns = {}
    for prefix in root.nsmap:
        separator = ':'
        nsprefix = prefix
        if prefix is None:
            nsprefix = ''
            separator = ''
        ns['xmlns{0}{1}'.format(separator, nsprefix)] = root.nsmap[prefix]
    return ns


def find_element_xpath(base, name, ns):
    return base.find('.//{%s}%s' % (ns['xmlns'], name))


def generate_tags_multiple_files(input_files, tag, ignore_tags, ns=None):
    """
    Calls xmltag generator for multiple files.
    """
    return itertools.chain.from_iterable([generate_xmltags(
        fn, tag, ignore_tags, ns) for fn in input_files])


def generate_tags_multiple_files_strings(input_files, ns, tag, ignore_tags):
    """
    Creates stringified xml output of elements with certain tag.
    """
    for el in generate_tags_multiple_files(input_files, tag, ignore_tags, ns):
        yield formatting.string_and_clear(el, ns)


def generate_xmltags(fn, returntag, ignore_tags, ns=None):
    """
    Base generator for percolator xml psm, peptide, protein output,
    as well as for mzML, mzIdentML.
    ignore_tags are the ones that are cleared when met by parser.
    """
    xmlns = create_namespace(ns)
    ns_ignore = ['{0}{1}'.format(xmlns, x) for x in ignore_tags]
    for ac, el in etree.iterparse(fn):
        if el.tag == '{0}{1}'.format(xmlns, returntag):
            yield el
        elif el.tag in ns_ignore:
            formatting.clear_el(el)


def get_element(fn, tag, ns=None):
    xmlns = create_namespace(ns)
    for ac, el in etree.iterparse(fn):
        if el.tag == '{0}{1}'.format(xmlns, tag):
            return el


def create_namespace(ns):
    if ns is None:
        return ''
    else:
        return '{%s}' % ns['xmlns']



from app.actions.prottable import create_empty as preparation
from app.readers import tsv as reader
from app.drivers.prottable.base import ProttableDriver
from app.drivers.options import prottable_options


class CreateEmptyDriver(ProttableDriver):
    outsuffix = '_prottable.txt'
    command = 'emptytable'
    commandhelp = ('Create protein table from PSM table containing no '
                   'quant data, resulting in one column with (master) '
                   'proteins only. Use --protcol if input PSMs '
                   'are not of standard protein grouped variety.')

    def __init__(self):
        super().__init__()
        self.infiletype = 'TSV PSM table (MSGF+)'

    def set_options(self):
        super().set_options()
        self.options.update(self.define_options(['proteincol']))

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.headertypes = []

    def initialize_input(self):
        self.pepheader = reader.get_tsv_header(self.fn)
        self.in_psms = reader.generate_tsv_psms(self.fn, self.pepheader)

    def set_feature_generator(self):
        if self.proteincol is not None:
            self.get_column_header_for_number(['proteincol'], self.pepheader)
        self.features = preparation.generate_master_proteins(self.in_psms,
                                                             self.proteincol)



from app.dataformats import mzidtsv as mzidtsvdata
from app.dataformats import prottable as prottabledata


def generate_master_proteins(psms, protcol):
    """Fed with a psms generator, this returns the master proteins present
    in the PSM table. PSMs with multiple master proteins are excluded."""
    master_proteins = {}
    if not protcol:
        protcol = mzidtsvdata.HEADER_MASTER_PROT
    for psm in psms:
        protacc = psm[protcol]
        if ';' in protacc:
            continue
        master_proteins[protacc] = 1
    if 'NA' in master_proteins:
        master_proteins.pop('NA')
    if '' in master_proteins:
        master_proteins.pop('')
    for protacc in master_proteins:
        yield {prottabledata.HEADER_PROTEIN: protacc}



from app.lookups.sqlite.base import ResultLookupInterface


# Indices that belong to positions of these features in output from
# function get_all_psms_proteingroups:
MASTER_INDEX = 1
PROTEIN_ACC_INDEX = 2
PEPTIDE_COUNT_INDEX = 3
PSM_COUNT_INDEX = 4
PROTEIN_SCORE_INDEX = 5
COVERAGE_INDEX = 6
EVIDENCE_LVL_INDEX = 7


class ProteinGroupDB(ResultLookupInterface):
    def add_tables(self):
        self.create_tables(['protein_coverage', 'protein_group_master',
                            'protein_group_content', 'psm_protein_groups'])

    def store_masters(self, allmasters, psm_masters):
        allmasters = ((x,) for x in allmasters)
        cursor = self.get_cursor()
        cursor.executemany(
            'INSERT INTO protein_group_master(protein_acc) VALUES(?)',
            allmasters)
        master_ids = self.get_master_ids()
        psms = ((psm_id, master_ids[master])
                for psm_id, masters in psm_masters.items()
                for master in masters)
        cursor.executemany(
            'INSERT INTO psm_protein_groups(psm_id, master_id) '
            'VALUES(?, ?)', psms)
        self.conn.commit()
        self.index_column('psm_pg_index', 'psm_protein_groups', 'master_id')
        self.index_column('psm_pg_psmid_index', 'psm_protein_groups', 'psm_id')

    def update_master_proteins(self, new_masters):
        cur = self.get_cursor()
        sql = 'UPDATE protein_group_master SET protein_acc=? WHERE master_id=?'
        cur.executemany(sql, new_masters)
        self.conn.commit()

    def get_master_ids(self):
        cur = self.get_cursor()
        cur.execute('SELECT protein_acc, master_id FROM protein_group_master')
        return {p_acc: master_id for (p_acc, master_id) in cur}

    def store_coverage(self, coverage):
        cursor = self.get_cursor()
        sql = ('INSERT INTO protein_coverage(protein_acc, coverage) '
               'VALUES(?, ?)')
        cursor.executemany(sql, coverage)
        self.conn.commit()
        self.index_column('cov_index', 'protein_coverage', 'protein_acc')

    def store_protein_group_content(self, protein_groups):
        cursor = self.get_cursor()
        cursor.executemany('INSERT INTO protein_group_content('
                           'protein_acc, master_id, peptide_count, '
                           'psm_count, protein_score) '
                           'VALUES(?, ?, ?, ?, ?)', protein_groups)
        self.conn.commit()

    def index_protein_group_content(self):
        self.index_column('pgc_master_index', 'protein_group_content',
                          'master_id')

    def get_all_psm_protein_relations(self):
        sql = 'SELECT psm_id, protein_acc FROM protein_psm'
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_allpsms_masters(self):
        sql = ('SELECT pgm.protein_acc, pp.psm_id FROM protein_group_master '
               'AS pgm JOIN protein_psm as pp USING(protein_acc)')
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_proteins_for_peptide(self, psm_id):
        """Returns list of proteins for a passed psm_id"""
        protsql = self.get_sql_select(['protein_acc'], 'protein_psm')
        protsql = '{0} WHERE psm_id=?'.format(protsql)
        cursor = self.get_cursor()
        proteins = cursor.execute(protsql, psm_id).fetchall()
        return [x[0] for x in proteins]

    def get_protpepmap_from_proteins(self, proteins):
        pepsql = self.get_sql_select(['protein_acc', 'psm_id'],
                                     'protein_psm',
                                     distinct=True)
        pepsql = '{0} WHERE protein_acc {1}'.format(
            pepsql, self.get_inclause(proteins))
        cursor = self.get_cursor()
        protpeps = cursor.execute(pepsql, proteins).fetchall()
        outmap = {}
        for protein, peptide in protpeps:
            try:
                outmap[protein].append(peptide)
            except KeyError:
                outmap[protein] = [peptide]
        return outmap

    def get_all_proteins_psms_seq(self):
        sql = ('SELECT p.protein_acc, ps.sequence, pp.psm_id, peps.sequence '
               'FROM proteins AS p '
               'JOIN protein_seq AS ps USING(protein_acc) '
               'JOIN protein_psm AS pp USING(protein_acc) '
               'JOIN psms AS psms USING(psm_id) '
               'JOIN peptide_sequences AS peps USING(pep_id)'
               )
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_protein_psm_records(self):
        sql = ('SELECT protein_acc, psm_id FROM protein_psm')
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def get_protein_group_candidates(self):
        sql = ('SELECT pgm.master_id, pgm.psm_id, pp.protein_acc, '
               'peps.sequence, p.score, pev.evidence_lvl, pc.coverage '
               'FROM psm_protein_groups AS pgm '
               'JOIN protein_psm AS pp USING(psm_id) '
               'JOIN psms AS p USING(psm_id) '
               'JOIN peptide_sequences AS peps USING(pep_id) '
               'LEFT OUTER JOIN protein_evidence AS pev '
               'ON pev.protein_acc=pp.protein_acc '
               'LEFT OUTER JOIN protein_coverage AS pc '
               'ON pc.protein_acc=pp.protein_acc '
               'ORDER BY pgm.master_id'
               )
        cursor = self.get_cursor()
        return cursor.execute(sql)

    def check_evidence_tables(self):
        """Returns True if there are records in evidence
        tables, otherwise returns False"""
        return self.check_table('protein_evidence') is not False

    def check_table(self, tablename):
        cursor = self.get_cursor()
        cursor.execute('SELECT * FROM {} LIMIT 10'.format(tablename))
        if len(cursor.fetchall()) > 0:
            return True
        return False

    def get_all_psms_proteingroups(self, evidence):
        fields = ['pr.rownr', 'pgm.protein_acc', 'pgc.protein_acc',
                  'pgc.peptide_count', 'pgc.psm_count', 'pgc.protein_score',
                  'pc.coverage']
        joins = [('psm_protein_groups', 'ppg', 'psm_id'),
                 ('protein_group_master', 'pgm', 'master_id'),
                 ('protein_group_content', 'pgc', 'master_id'),
                 ]
        join_sql = '\n'.join(['JOIN {0} AS {1} USING({2})'.format(
            j[0], j[1], j[2]) for j in joins])
        specialjoin = []
        if evidence:
            specialjoin = [('protein_evidence', 'pev')]
            fields.append('pev.evidence_lvl')
        specialjoin.append(('protein_coverage', 'pc'))
        specialjoin = '\n'.join(['JOIN {0} AS {1} ON '
                                 'pgc.protein_acc={1}.protein_acc'.format(
                                     j[0], j[1]) for j in specialjoin])
        join_sql = '{} {}'.format(join_sql, specialjoin)
        sql = 'SELECT {0} FROM psmrows AS pr {1} ORDER BY pr.rownr'.format(
            ', '.join(fields), join_sql)
        cursor = self.get_cursor()
        return cursor.execute(sql)



from app.readers import tsv as tsvreader
from app.dataformats import mzidtsv as mzidtsvdata


def merge_mzidtsvs(fns, header):
    for fn in fns:
        if header != tsvreader.get_tsv_header(fn):
            raise RuntimeError('Headers of TSV files to concatenate are '
                               'not identical')
    for psm in tsvreader.generate_tsv_lines_multifile(fns, header):
        yield psm


def get_splitcolnr(header, bioset, splitcol):
    """Returns column nr on which to split PSM table. Chooses from flags
    given via bioset and splitcol"""
    if bioset:
        return header.index(mzidtsvdata.HEADER_SETNAME)
    elif splitcol is not None:
        return splitcol - 1
    else:
        raise RuntimeError('Must specify either --bioset or --splitcol')


def generate_psms_split(fn, oldheader, bioset, splitcol):
    """Loops PSMs and outputs dictionaries passed to writer. Dictionaries
    contain the PSMs and info to which split pool the
    respective PSM belongs"""
    try:
        splitcolnr = get_splitcolnr(oldheader, bioset, splitcol)
    except IndexError:
        raise RuntimeError('Cannot find bioset header column in '
                           'input file {}, though --bioset has '
                           'been passed'.format(fn))
    for psm in tsvreader.generate_tsv_psms(fn, oldheader):
        yield {'psm': psm,
               'split_pool': psm[oldheader[splitcolnr]]
               }



#!/usr/bin/python3

import sys

from app.drivers.mzidtsv import (spectra, percolator, proteingrouping,
                                 prot2gene, quant, splitmerge,
                                 filter_confidence, isonormalize)
from app.drivers import startup


def main():
    drivers = [spectra.TSVSpectraDriver(),
               percolator.MzidPercoTSVDriver(),
               splitmerge.MzidTSVConcatenateDriver(),
               splitmerge.MzidTSVSplitDriver(),
               filter_confidence.ConfidenceFilterDriver(),
               quant.TSVQuantDriver(),
               proteingrouping.ProteinGroupDriver(),
               prot2gene.TSVGeneFromProteinDriver(),
               isonormalize.PSMIsoquantRatioDriver(),
               isonormalize.PSMIsoquantNormalizeDriver(),
               ]
    startup.start_msstitch(drivers, sys.argv)



from numpy import polyfit
from math import log

from app.readers import tsv as reader
from app.dataformats import peptable as peptabledata


def recalculate_qvals_linear_model(fn, scorecol, qvalcol, qvalthreshold):
    slope, intercept = fit_linear_model(fn, scorecol, qvalcol, qvalthreshold)
    for peptide in reader.generate_tsv_peptides(fn):
        outpeptide = {k: v for k, v in peptide.items()}
        score = float(outpeptide[scorecol])
        qval = 10 ** (slope * score + intercept)
        outpeptide[peptabledata.HEADER_QVAL_MODELED] = str(qval)
        yield outpeptide


def fit_linear_model(fn, scorecol, qvalcol, qvalthreshold):
    pepq = []
    pepscore = []
    for peptide in reader.generate_tsv_peptides(fn):
        qval = float(peptide[qvalcol])
        if qval > qvalthreshold:
            pepq.append(log(qval, 10))
            pepscore.append(float(peptide[scorecol]))
    slope, intercept = polyfit(pepscore, pepq, deg=1)
    print('Fitted linear model through qvalues (above {}) vs score. '
          'Slope={}, intercept={}'.format(qvalthreshold, slope, intercept))
    return slope, intercept



DB_STORE_CHUNK = 500000


def create_spectra_lookup(lookup, fn_spectra):
    """Stores all spectra rt, injection time, and scan nr in db"""
    to_store = []
    mzmlmap = lookup.get_mzmlfile_map()
    for fn, spectrum in fn_spectra:
        spec_id = '{}_{}'.format(mzmlmap[fn], spectrum['scan'])
        mzml_rt = round(float(spectrum['rt']), 12)
        mzml_iit = round(float(spectrum['iit']), 12)
        mz = float(spectrum['mz'])
        to_store.append((spec_id, mzmlmap[fn], spectrum['scan'],
                         spectrum['charge'], mz, mzml_rt, mzml_iit))
        if len(to_store) == DB_STORE_CHUNK:
            lookup.store_mzmls(to_store)
            to_store = []
    lookup.store_mzmls(to_store)
    lookup.index_mzml()



import os
import sys

from app.lookups import base as lookups
from app.drivers.options import shared_options


class BaseDriver(object):
    def __init__(self):
        self.lookupfn = None
        self.infiletype = ''

    def set_lookup(self):
        if self.lookupfn is not None and hasattr(self, 'lookuptype'):
            self.lookup = lookups.get_lookup(self.lookupfn, self.lookuptype)
        else:
            self.lookup = None

    def set_options(self):
        self.options = self.define_options(['fn', 'outdir', 'outfile'], {})
        self.options['-i']['help'] = self.options['-i']['help'].format(
            self.infiletype)

    def get_commandhelp(self):
        return self.commandhelp

    def get_options(self):
        return self.options.values()

    def define_options(self, names, parser_options=None):
        """Given a list of option names, this returns a list of dicts
        defined in all_options and self.shared_options. These can then
        be used to populate the argparser with"""
        def copy_option(options, name):
            return {k: v for k, v in options[name].items()}
        if parser_options is None:
            parser_options = {}
        options = {}
        for name in names:
            try:
                option = copy_option(parser_options, name)
            except KeyError:
                option = copy_option(shared_options, name)
            try:
                options.update({option['clarg']: option})
            except TypeError:
                options.update({option['clarg'][0]: option})
        return options

    def parse_input(self, **kwargs):
        for option in self.get_options():
            opt_argkey = option['driverattr']
            opt_val = kwargs.get(opt_argkey)
            if 'type' in option and option['type'] == 'pick':
                if not option.get('required') and not opt_val:
                    pass
                else:
                    try:
                        assert opt_val in option['picks']
                    except AssertionError:
                            print('Option {} should be one of [{}]'.format(
                                option['clarg'], ','.join(option['picks'])))
                            sys.exit(1)
            setattr(self, opt_argkey, opt_val)
        if self.outdir is None:
            self.outdir = os.getcwd()
        if self.outfile is not None:
            if not os.path.isabs(self.outfile):
                self.outfile = os.path.join(self.outdir, self.outfile)

    def start(self, **kwargs):
        self.parse_input(**kwargs)
        self.set_lookup()
        self.run()

    def finish(self):
        """Cleans up after work"""
        pass

    def create_outfilepath(self, fn, suffix=None):
        if self.outfile is None:
            basefn = os.path.basename(fn)
            outfn = os.path.join(self.outdir, basefn + suffix)
        else:
            outfn = self.outfile
        return outfn

    def number_to_headerfield(self, columnr, header):
        return header[int(columnr) - 1]

    def get_column_header_for_number(self, column_var_names, header=False):
        """This function subtracts 1 from inputted column number to comply
        with programmers counting (i.e. from 0, not from 1). For TSV data."""
        if not header:
            header = self.oldheader
        for col in column_var_names:
            value = getattr(self, col)
            if not value or value is None:
                continue
            setattr(self, col, self.number_to_headerfield(value, header))

    def get_commandname(self):
        return self.command

    def get_fastadelim_genefield(self, delimtype, genefield):
        if delimtype is not None:
            fastadelim = {'tab': '\t',
                          'semicolon': ';', 'pipe': '|'}[delimtype]
        else:
            fastadelim = None
        if genefield is not None:
            genefield -= 1
        return fastadelim, genefield



from app.drivers.pycolator import base
from app.actions.pycolator import splitmerge as preparation
from app.readers import pycolator as readers
from app.drivers.options import pycolator_options


class SplitDriver(base.PycolatorDriver):
    outfile = None

    def run(self):
        self.set_filter_types()
        for filter_type, suffix in self.filter_types:
            self.prepare()
            self.set_features(filter_type)
            self.outsuffix = suffix
            self.write()
        self.finish()

    def set_options(self):
        """Since splitdriver splits into multiple files we cannot set an
        output file"""
        super().set_options()
        del(self.options['-o'])

    def set_features(self, filter_type):
        """Calls splitter to split percolator output into target/decoy
        elements.
        Writes two new xml files with features. Currently only psms and
        peptides. Proteins not here, since one cannot do protein inference
        before having merged and remapped multifraction data anyway.
        """
        elements_to_split = {'psm': self.allpsms, 'peptide': self.allpeps}
        self.features = self.splitfunc(elements_to_split, self.ns, filter_type)


class SplitTDDriver(SplitDriver):
    command = 'splittd'
    commandhelp = ('Splits target and decoy data, producing 2 output files')

    def set_filter_types(self):
        self.filter_types = [('target', '_target.xml'),
                             ('decoy', '_decoy.xml')]

    def set_features(self, filter_type):
        self.splitfunc = preparation.split_target_decoy
        super().set_features(filter_type)


class SplitProteinDriver(SplitDriver):
    command = 'splitprotein'
    commandhelp = ('Splits input XML into multiple files depending based on '
                   'the protein headers specified. Each header class gets '
                   'its own output file')

    def set_filter_types(self):
        maxdigits = len(str(len(self.protheaders)))
        self.filter_types = [(headers, '_h{i:0{dig}d}.xml'.format(
            i=ix, dig=maxdigits))
            for ix, headers in enumerate(self.protheaders)]

    def set_features(self, filter_type):
        self.splitfunc = preparation.split_protein_header_id_type
        super().set_features(filter_type)

    def set_options(self):
        super().set_options()
        options = self.define_options(['protheaders'], pycolator_options)
        self.options.update(options)


class MergeDriver(base.PycolatorDriver):
    """Base class for merging multiple percolator fractions under different
    sorts of filtering. It writes a single percolator out xml from
    multiple fractions.
    Namespace and static xml come from first percolator file.
    Make sure fractions are from same percolator run."""
    outsuffix = '_merged.xml'
    command = 'merge'
    commandhelp = 'Merges percolator xml files, nothing else.'

    def parse_input(self, **kwargs):
        super().parse_input(**kwargs)
        self.mergefiles = self.fn[:]
        self.fn = self.fn[0]

    def set_options(self):
        super().set_options()
        options = self.define_options(['multifiles'], pycolator_options)
        self.options.update(options)

    def prepare(self):
        self.ns, self.static_xml = self.prepare_percolator_output(self.fn)

    def set_features(self):
        """"Merge all psms and peptides"""
        allpsms_str = readers.generate_psms_multiple_fractions_strings(
            self.mergefiles, self.ns)
        allpeps = preparation.merge_peptides(self.mergefiles, self.ns)
        self.features = {'psm': allpsms_str, 'peptide': allpeps}



from app.drivers import base
from app.readers import tsv as tsvreader
from app.writers import mzidtsv as writers
from app.drivers.options import mzidtsv_options


class MzidTSVDriver(base.BaseDriver):
    def __init__(self):
        super().__init__()
        self.infiletype = 'TSV PSM table (MSGF+)'
        self.parser_options = mzidtsv_options

    def run(self):
        if type(self.fn) == list:
            self.first_infile = self.fn[0]
        else:
            self.first_infile = self.fn
        self.oldheader = tsvreader.get_tsv_header(self.first_infile)
        self.get_psms()
        self.write()
        self.finish()

    def write(self):
        outfn = self.create_outfilepath(self.first_infile, self.outsuffix)
        writers.write_mzid_tsv(self.header, self.psms, outfn)
